import logging
from pathlib import Path

from PySide6.QtCore import Signal, QThread

from freemocap.utilities.download_sample_data import download_dataset

logger = logging.getLogger(__name__)


class DownloadDataThreadWorker(QThread):
    finished = Signal(str)
    in_progress = Signal(str)

    def __init__(self, dataset_name: str, parent=None):
        super().__init__(parent=parent)
        logger.debug("Initializing download sample data thread worker")
        self.dataset_name = dataset_name

    def run(self):
        logger.debug("Starting sample data download")

        try:
            downloaded_data_path = download_dataset(self.dataset_name)
            if Path(downloaded_data_path).exists():
                logger.debug(f"Data successfully downloaded to {downloaded_data_path}")
                self.finished.emit(downloaded_data_path)
            else:
                logger.error(f"Could not find downloaded data at {downloaded_data_path}")
                raise FileNotFoundError(f"Could not find downloaded data at {downloaded_data_path}")

        except Exception as e:  # noqa
            logger.exception(e)
            logger.error("Error downloading sample data")
            raise e



import logging
from pathlib import Path
from typing import Tuple, Union

import numpy as np

from freemocap.core_processes.capture_volume_calibration.save_3d_data_to_npy import (
    save_3d_data_to_npy,
)
from freemocap.utilities.geometry.project_3d_data_to_z_plane import project_3d_data_to_z_plane
from freemocap.utilities.geometry.rotate_by_90_degrees_around_x_axis import rotate_by_90_degrees_around_x_axis

logger = logging.getLogger(__name__)


def process_single_camera_skeleton_data(
        input_image_data_frame_marker_xyz: np.ndarray,
        raw_data_folder_path: Union[str, Path],
        file_prefix: str = "",
        project_to_z_plane: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    skeleton_reprojection_error_fr_mar = np.zeros(input_image_data_frame_marker_xyz.shape[0:2])
    skeleton_reprojection_error_cam_fr_mar = np.zeros(
        (1, input_image_data_frame_marker_xyz.shape[0], input_image_data_frame_marker_xyz.shape[1])
    )
    if project_to_z_plane:
        logger.info("Altering image-derived 3d data to resemble multi-camera reconstructed data.")

        raw_skel3d_frame_marker_xyz = project_3d_data_to_z_plane(
            skel3d_frame_marker_xyz=input_image_data_frame_marker_xyz
        )
    else:
        raw_skel3d_frame_marker_xyz = input_image_data_frame_marker_xyz

    save_3d_data_to_npy(
        data3d_numFrames_numTrackedPoints_XYZ=raw_skel3d_frame_marker_xyz,
        data3d_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_fr_mar,
        data3d_numCams_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_cam_fr_mar,
        path_to_folder_where_data_will_be_saved=raw_data_folder_path,
        processing_level="raw",
        file_prefix=file_prefix,
    )
    rotated_raw_skel3d_frame_marker_xyz = rotate_by_90_degrees_around_x_axis(raw_skel3d_frame_marker_xyz)

    return rotated_raw_skel3d_frame_marker_xyz, skeleton_reprojection_error_fr_mar



from pathlib import Path
import numpy as np
from time import perf_counter_ns

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.freemocap_anipose import (
    CameraGroup,
    AniposeCharucoBoard,
)
from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)
from freemocap.utilities.get_video_paths import get_video_paths


if __name__ == "__main__":
    # TODO: do all this setup with recording_info object
    recording_folder = Path("/Users/philipqueen/freemocap_data/recording_sessions/freemocap_test_data/")
    synchronized_folder = recording_folder / "synchronized_videos"

    video_paths = get_video_paths(synchronized_folder)
    list_of_camera_names = [video_path.stem for video_path in video_paths]
    video_paths_list_of_list_of_strings = [[str(this_path)] for this_path in video_paths]

    calibration = CameraGroup.from_names(list_of_camera_names)
    charuco_square_size = 39

    calibration.metadata["charuco_square_size"] = charuco_square_size
    calibration.metadata["charuco_board_object"] = str(CharucoBoardDefinition)
    calibration.metadata["path_to_recorded_videos"] = str(synchronized_folder)
    calibration.metadata["date_time_calibrated"] = str(np.datetime64("now"))

    anipose_charuco_board = AniposeCharucoBoard(
        7,
        6,
        square_length=charuco_square_size,  # mm
        marker_length=charuco_square_size * 0.8,
        marker_bits=4,
        dict_size=250,
    )

    num_trials = 5
    print("Benchmarking calibrate_videos:\n")
    times = []
    for i in range(num_trials):
        start = perf_counter_ns()
        calibration.calibrate_videos(video_paths_list_of_list_of_strings, anipose_charuco_board)
        end = perf_counter_ns()
        times.append((end - start) / 1e9)
        print(f"This round calibration time is: {(end - start) / 1e9} seconds")
    print(f"\tAverage calibration time across trials is: {np.mean(times)} seconds over {num_trials} runs")



import logging
from typing import Dict, Tuple

import numpy as np

from freemocap.data_layer.skeleton_models.segments import SegmentAnthropometry
from freemocap.data_layer.skeleton_models.skeleton import Skeleton

logger = logging.getLogger(__name__)


def calculate_all_segments_com(
        segment_positions: Dict[str, Dict[str, np.ndarray]], center_of_mass_definitions: Dict[str, SegmentAnthropometry]
) -> Dict[str, np.ndarray]:
    """
    Calculates the center of mass (COM) for each segment based on provided segment positions
    and anthropometric data.

    Parameters:
    - segment_positions: A dictionary where each key is a segment name, and the value is another
      dictionary with 'proximal' and 'distal' keys containing the respective marker positions
      as numpy arrays.
    - center_of_mass_definitions: A dictionary containing anthropometric information for each segment.
      Each key is a segment name with a value that is a dictionary, which includes the
      'segment_com_length' key representing the percentage distance from the proximal marker
      to the segment's COM.

    Returns:
    - A dictionary where each key is a segment name, and the value is the calculated COM
      position as a numpy array.
    """

    segment_com_data = {}

    for segment_name, segment_info in center_of_mass_definitions.items():
        proximal = segment_positions[segment_name]["proximal"]
        distal = segment_positions[segment_name]["distal"]

        com_length = segment_info.segment_com_length

        segment_com = proximal + (distal - proximal) * com_length

        segment_com_data[segment_name] = segment_com

    return segment_com_data


def calculate_total_body_center_of_mass(
        segment_center_of_mass_data: Dict[str, np.ndarray],
        center_of_mass_definitions: Dict[str, SegmentAnthropometry],
        num_frames: int,
) -> np.ndarray:
    """
    Calculates the total body center of mass for each frame based on segment COM positions and anthropometric data.

    Parameters:
    - segment_com_data: A dictionary with segment names as keys and COM positions as values for each frame.
    - center_of_mass_definitions: A dictionary containing segment mass percentages.
    - num_frames: The number of frames in the data.

    Returns:
    - A numpy array containing the position of the total body center of mass for each frame.
    """
    total_body_com = np.zeros((num_frames, 3))

    for segment_name, segment_info in center_of_mass_definitions.items():

        segment_com = segment_center_of_mass_data.get(segment_name)
        if segment_com is None:
            raise ValueError(f"Segment {segment_name} not found in segment center of mass data.")

        segment_mass_percentage = segment_info.segment_com_percentage

        total_body_com += segment_com * segment_mass_percentage

    return total_body_com


def get_all_segment_markers(skeleton: Skeleton) -> Dict[str, Dict[str, np.ndarray]]:
    """
    Retrieves the proximal and distal marker positions for all segments within a skeleton.

    Parameters:
    - skeleton: The Skeleton instance containing segment information and marker data.

    Returns:
    - A dictionary where each key is a segment name, and the value is another dictionary
      with 'proximal' and 'distal' keys containing the respective marker positions as numpy arrays.
    """
    segment_positions = {}

    if skeleton.segments is None:
        raise ValueError("Segments must be defined before getting segment markers.")

    for segment_name in skeleton.segments.keys():
        segment_positions[segment_name] = skeleton.get_segment_markers(segment_name)

    return segment_positions


def merge_segment_com_data(segment_com_data: Dict[str, np.ndarray]) -> np.ndarray:
    """
    Merges the center of mass data from multiple segments into a single array.

    Parameters:
    - segment_com_data: A dictionary where each key is a segment name and the value is the center of mass data for that segment.

    Returns:
    - A numpy array containing the merged center of mass data.
    """
    com_data_list = list(segment_com_data.values())

    return np.stack(com_data_list, axis=1)


def calculate_center_of_mass_from_skeleton(skeleton: Skeleton) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculates the segment and total body center of mass for a skeleton based on anthropometric data.

    Parameters:
    - skeleton: The Skeleton instance containing marker data and segment information.
    - anthropometric_data: A dictionary containing segment mass percentages

    Returns:
    - A tuple containing the segment center of mass data and the total body center of mass.
    """
    if skeleton.center_of_mass_definitions is None:
        raise ValueError("Segment center of mass definitions must be defined before calculating center of mass.")

    if skeleton.segments is None:
        raise ValueError("Segments must be defined before calculating center of mass.")

    segment_3d_positions = get_all_segment_markers(skeleton)  # TODO: Should this be a method of the skeleton model?
    segment_com_data = calculate_all_segments_com(segment_3d_positions, skeleton.center_of_mass_definitions)
    total_body_com = calculate_total_body_center_of_mass(
        segment_com_data, skeleton.center_of_mass_definitions, skeleton.num_frames
    )

    merged_segment_com_data = merge_segment_com_data(segment_com_data)

    return merged_segment_com_data, total_body_com



import logging
import os
import threading
from pathlib import Path
from typing import Callable, Optional, Union

from PySide6.QtCore import Qt, Slot, Signal, QObject
from PySide6.QtGui import QDoubleValidator
from PySide6.QtWidgets import (
    QFileDialog,
    QFormLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
    QWidget,
    QHBoxLayout,
    QLayout,
    QCheckBox,
    QComboBox,
)

from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import CHARUCO_BOARDS
from freemocap.data_layer.recording_models.recording_info_model import RecordingInfoModel
from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState, load_gui_state, save_gui_state
from freemocap.gui.qt.widgets.calibration_guide_link_qlabel import CalibrationGuideLinkQLabel
from freemocap.gui.qt.widgets.groundplane_failure_dialog import GroundPlaneCalibrationFailedDialog
from freemocap.gui.qt.workers.anipose_calibration_thread_worker import (
    AniposeCalibrationThreadWorker,
)
from freemocap.system.paths_and_filenames.path_getters import (
    get_gui_state_json_path,
    get_last_successful_calibration_toml_path,
)

logger = logging.getLogger(__name__)


class CalibrationControlPanel(QWidget):
    control_panel_calibration_updated = Signal()

    def __init__(
        self,
        get_active_recording_info: Callable[..., Union[RecordingInfoModel, Path]],
        kill_thread_event: threading.Event,
        gui_state: GuiState,
        parent: Optional[QObject] = None,
    ):
        super().__init__(parent=parent)
        self.gui_state = gui_state
        self._has_a_toml_path = False
        self._calibration_toml_path = None
        self._get_active_recording_info = get_active_recording_info
        self._kill_thread_event = kill_thread_event

        self.parent = parent

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._calibration_toml_path = None
        self._kill_thread_event = kill_thread_event

        self._anipose_calibration_frame_worker = None
        self._user_selected_toml_path = None

        self._selected_calibration_toml_label = QLabel("")
        self._selected_calibration_toml_label.setWordWrap(True)

        self._layout.addWidget(self._selected_calibration_toml_label)

        self._radio_button_layout = self._create_radio_button_layout()
        self._radio_button_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._layout.addLayout(self._radio_button_layout)
        # self._layout.addStretch()

    @property
    def calibration_toml_path(self) -> str:
        return self._calibration_toml_path

    @property
    def radio_button_layout(self) -> QLayout:
        return self._radio_button_layout

    def _create_radio_button_layout(self):
        radio_button_form_layout = QFormLayout()

        radio_button_form_layout.addRow(self._create_use_most_recent_calibration_radio_button())

        radio_button_form_layout.addRow(self._create_load_calibration_from_file_radio_button())

        radio_button_form_layout.addRow(self._add_calibrate_from_active_recording_radio_button())

        radio_button_form_layout.addRow(self._create_charuco_options_layout())

        self.update_calibration_toml_path()
        return radio_button_form_layout

    def _create_load_calibration_from_file_radio_button(self) -> QLayout:
        vbox_layout = QVBoxLayout()
        radio_button_push_button_layout = QHBoxLayout()
        vbox_layout.addLayout(radio_button_push_button_layout)

        self._load_calibration_from_file_radio_button = QRadioButton("Load calibration from file")
        radio_button_push_button_layout.addWidget(self._load_calibration_from_file_radio_button)
        self._load_calibration_from_file_radio_button.toggled.connect(self._handle_load_calibration_from_file_toggled)

        self._load_calibration_toml_dialog_button = QPushButton("Load TOML...")
        radio_button_push_button_layout.addWidget(self._load_calibration_toml_dialog_button)
        # self._load_calibration_toml_dialog_button.setStyleSheet("font-size: 10pt;")
        self._load_calibration_toml_dialog_button.clicked.connect(self.open_load_camera_calibration_toml_dialog)
        self._load_calibration_toml_dialog_button.setEnabled(False)

        return vbox_layout

    def update_calibration_toml_path(self, toml_path: str = None):
        active_recording_info = self._get_active_recording_info()
        if toml_path is None and active_recording_info is not None:
            if active_recording_info.single_video_check:
                pass

            elif self._load_calibration_from_file_radio_button.isChecked():
                if self._user_selected_toml_path is not None:
                    toml_path = self._user_selected_toml_path
                else:
                    toml_path = self._check_active_recording_for_calibration_toml()

            elif self._use_most_recent_calibration_radio_button.isChecked():
                toml_path = get_last_successful_calibration_toml_path()

            elif self._calibrate_from_active_recording_radio_button.isChecked():
                toml_path = self._check_active_recording_for_calibration_toml()

            else:  # no button checked -> initialize
                toml_path = self._check_active_recording_for_calibration_toml()
                if toml_path is not None:
                    self._load_calibration_from_file_radio_button.setChecked(True)
                else:
                    toml_path = get_last_successful_calibration_toml_path()
                    self._use_most_recent_calibration_radio_button.setChecked(True)

        self._calibration_toml_path = toml_path

        if self._calibration_toml_path is not None:
            logger.debug(f"Setting calibration TOML path: {self._calibration_toml_path}")
            self._show_selected_calibration_toml_path(self._calibration_toml_path)
        elif active_recording_info is None:
            self._show_selected_calibration_toml_path("-No Active Recording-")
        elif active_recording_info.single_video_check:
            self._show_selected_calibration_toml_path("-Single Video Recording, No Calibration Needed-")
        else:
            self._show_selected_calibration_toml_path("-Calibration File Not Found-")

    def _bring_up_groundplane_calibration_failed_dialog(self, message: str):
        failure_dialog = GroundPlaneCalibrationFailedDialog(message=message)
        failure_dialog.exec()

    def _add_calibrate_from_active_recording_radio_button(self):
        hbox = QHBoxLayout()

        self._calibrate_from_active_recording_radio_button = QRadioButton("Calibrate from Active Recording")
        hbox.addWidget(self._calibrate_from_active_recording_radio_button)

        self._calibrate_from_active_recording_radio_button.toggled.connect(
            self._handle_calibrate_from_active_recording_toggled
        )

        self._calibrate_from_active_recording_button = QPushButton("Run calibration...")
        hbox.addWidget(self._calibrate_from_active_recording_button)

        self._calibrate_from_active_recording_button.setEnabled(False)
        self._calibrate_from_active_recording_button.clicked.connect(self.calibrate_from_active_recording)

        return hbox

    def _create_charuco_options_layout(self):
        vbox = QVBoxLayout()

        hbox1 = QHBoxLayout()
        hbox1.setAlignment(Qt.AlignmentFlag.AlignLeft)

        # Square size form layout
        self._charuco_square_size_form_layout = self._create_charuco_square_size_form_layout()
        hbox1.addLayout(self._charuco_square_size_form_layout)

        hbox1.addSpacing(16)

        # Board dropdown + label
        self._board_dropdown_label = QLabel("Charuco Board:")
        self._board_dropdown_label.setStyleSheet("QLabel { font-size: 12px; }")
        self._board_dropdown_label.setEnabled(False)
        self._board_dropdown = self._create_board_dropdown()
        self._board_dropdown.currentTextChanged.connect(self._on_charuco_board_dropdown_changed)
        hbox1.addWidget(self._board_dropdown_label)
        hbox1.addWidget(self._board_dropdown)
        vbox.addLayout(hbox1)

        # Groundplane checkbox row
        hbox2 = QHBoxLayout()
        hbox2.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self._use_charuco_as_groundplane_checkbox = QCheckBox(
            "Use initial Charuco board position as groundplane origin"
        )
        self._calibration_guide_link = CalibrationGuideLinkQLabel(parent=self)
        self._calibration_guide_link.setVisible(False)

        self._use_charuco_as_groundplane_checkbox.setStyleSheet("QCheckBox { font-size: 12px; }")
        self._use_charuco_as_groundplane_checkbox.setToolTip(
            "Set the Charuco board's coordinate system as the global origin"
        )
        self._use_charuco_as_groundplane_checkbox.setChecked(False)
        self._use_charuco_as_groundplane_checkbox.setEnabled(False)
        self._use_charuco_as_groundplane_checkbox.setVisible(False)
        self._use_charuco_as_groundplane_checkbox.stateChanged.connect(
            self._on_use_charuco_groundplane_checkbox_changed
        )

        hbox2.addWidget(self._use_charuco_as_groundplane_checkbox)

        hbox2.addWidget(self._calibration_guide_link)
        vbox.addLayout(hbox2)

        # Hide by default
        self._set_charuco_square_size_form_layout_visibility(False)
        self._set_charuco_board_dropdown_visibility(False)

        return vbox

    def _set_charuco_board_dropdown_visibility(self, visible: bool) -> None:
        self._board_dropdown.setEnabled(visible)
        self._board_dropdown_label.setEnabled(visible)
        self._board_dropdown.setVisible(visible)
        self._board_dropdown_label.setVisible(visible)

    def _create_use_most_recent_calibration_radio_button(self):
        self._use_most_recent_calibration_radio_button = QRadioButton("Use most recent calibration")

        self._use_most_recent_calibration_radio_button.toggled.connect(self._handle_use_most_recent_calibration_toggled)

        self._use_most_recent_calibration_radio_button.setToolTip(get_last_successful_calibration_toml_path())

        return self._use_most_recent_calibration_radio_button

    def update_calibrate_from_active_recording_button_text(self):
        active_recording_info = self._get_active_recording_info()
        if active_recording_info is None:
            active_path_str = "- No active recording selected -"
        else:
            if not active_recording_info.synchronized_videos_status_check:
                active_path_str = f"Recording: {active_recording_info.name} has no synchronized videos!"
            elif active_recording_info.single_video_check:
                active_path_str = "Single Video Recording: No calibration needed"
            else:
                active_path_str = f"Calibrate from Recording: {self._get_active_recording_info().name}"

        self._calibrate_from_active_recording_button.setToolTip(active_path_str)

    def _handle_use_most_recent_calibration_toggled(self, checked):
        self.update_calibration_toml_path()

    def _handle_load_calibration_from_file_toggled(self, checked):
        if checked:
            self._load_calibration_toml_dialog_button.setEnabled(True)
        else:
            self._load_calibration_toml_dialog_button.setEnabled(False)

        self.update_calibration_toml_path()

    def _handle_calibrate_from_active_recording_toggled(self, checked):
        active_recording_info = self._get_active_recording_info()
        self.update_calibrate_from_active_recording_button_text()
        self.update_calibration_toml_path()

        if checked and active_recording_info is not None and active_recording_info.synchronized_videos_status_check:
            self._calibrate_from_active_recording_button.setEnabled(True)
            self._set_charuco_square_size_form_layout_visibility(True)
            self._use_charuco_as_groundplane_checkbox.setEnabled(True)
            self._use_charuco_as_groundplane_checkbox.setVisible(True)
            self._set_charuco_board_dropdown_visibility(True)
            self._calibration_guide_link.setVisible(True)
        else:
            self._calibrate_from_active_recording_button.setEnabled(False)
            self._set_charuco_square_size_form_layout_visibility(False)
            self._use_charuco_as_groundplane_checkbox.setEnabled(False)
            self._use_charuco_as_groundplane_checkbox.setVisible(False)
            self._set_charuco_board_dropdown_visibility(False)
            self._calibration_guide_link.setVisible(False)

    def _set_charuco_square_size_form_layout_visibility(self, visible: bool) -> None:
        # Directly control the widgets using the stored references
        self._charuco_square_size_label.setEnabled(visible)
        self._charuco_square_size_label.setVisible(visible)
        self._charuco_square_size_line_edit.setEnabled(visible)
        self._charuco_square_size_line_edit.setVisible(visible)

    def open_load_camera_calibration_toml_dialog(self) -> str:
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)

        calibration_toml_path_selection = dialog.getOpenFileName(
            self,
            "Select 'toml' containing camera calibration info",
            str(self._get_active_recording_info().path),
            "Camera Calibration TOML (*.toml)",
        )
        if len(calibration_toml_path_selection) > 0:
            self._user_selected_toml_path = calibration_toml_path_selection[0]
            logger.info(f"User selected camera calibration toml path:{self._user_selected_toml_path}")
            self._show_selected_calibration_toml_path(self._user_selected_toml_path)
            self.update_calibration_toml_path(toml_path=self._user_selected_toml_path)
            return self._user_selected_toml_path

    def _show_selected_calibration_toml_path(self, calibration_toml_path_str: str):
        self._calibration_toml_path = calibration_toml_path_str
        path = calibration_toml_path_str.replace(os.sep, "/ ")
        self._selected_calibration_toml_label.setText(path)
        self._selected_calibration_toml_label.show()

    def _create_charuco_square_size_form_layout(self) -> QHBoxLayout:
        charuco_square_size_layout = QHBoxLayout()

        self._charuco_square_size_label = QLabel("Charuco square size (mm)")
        self._charuco_square_size_label.setStyleSheet("QLabel { font-size: 12px; }")

        # Ensure the label has enough width to show all text
        self._charuco_square_size_label.setMinimumWidth(200)  # Adjust this value as needed
        # OR use this to automatically size based on content:
        # self._charuco_square_size_label.adjustSize()

        self._charuco_square_size_line_edit = QLineEdit()
        self._charuco_square_size_line_edit.setValidator(QDoubleValidator())
        self._charuco_square_size_line_edit.setFixedWidth(65)
        self._charuco_square_size_line_edit.setText(str(self.gui_state.charuco_square_size))
        self._charuco_square_size_line_edit.setToolTip(
            "The length of one of the edges of the black squares in the calibration board in mm"
        )
        self._charuco_square_size_line_edit.textEdited.connect(self._on_charuco_square_size_line_edit_changed)

        # Add widgets to horizontal layout
        charuco_square_size_layout.addWidget(self._charuco_square_size_label)
        charuco_square_size_layout.addStretch()
        charuco_square_size_layout.addWidget(self._charuco_square_size_line_edit)
        charuco_square_size_layout.addStretch()
        return charuco_square_size_layout

    def _create_board_dropdown(self) -> QComboBox:
        board_dropdown = QComboBox()
        board_dropdown.setToolTip("Select the Charuco board to use for calibration")
        board_dropdown.setFixedWidth(130)
        board_dropdown.setStyleSheet("QComboBox { font-size: 12px; }")
        board_dropdown.setEnabled(False)
        board_dropdown.setEditable(False)
        board_dropdown.addItems(list(CHARUCO_BOARDS.keys()))
        return board_dropdown

    def _on_charuco_square_size_line_edit_changed(self):
        try:
            self.gui_state.charuco_square_size = float(self._charuco_square_size_line_edit.text())
        except ValueError:
            return
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())
        self.control_panel_calibration_updated.emit()

    def _on_charuco_board_dropdown_changed(self):
        selected_board_name = self._board_dropdown.currentText()
        self.gui_state.charuco_board_name = selected_board_name
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())
        self.control_panel_calibration_updated.emit()

    def _on_use_charuco_groundplane_checkbox_changed(self):
        self.gui_state.use_charuco_as_groundplane = self._use_charuco_as_groundplane_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())
        self.control_panel_calibration_updated.emit()

    @Slot()
    def charuco_option_updated(self):
        self.gui_state = load_gui_state(file_pathstring=get_gui_state_json_path())
        self._board_dropdown.setCurrentText(self.gui_state.charuco_board_name)
        self._charuco_square_size_line_edit.setText(str(self.gui_state.charuco_square_size))
        self._use_charuco_as_groundplane_checkbox.setChecked(self.gui_state.use_charuco_as_groundplane)

    @Slot(str)
    def _log_calibration_progress_callbacks(self, message: str):
        logger.info(message)

    def calibrate_from_active_recording(
        self,
        charuco_square_size_mm: float = None,
        use_charuco_as_groundplane: bool = None,
        charuco_board_name: str = None,
    ):
        if not charuco_square_size_mm:
            charuco_square_size_mm = float(self._charuco_square_size_line_edit.text())

        if not use_charuco_as_groundplane:
            use_charuco_as_groundplane = self._use_charuco_as_groundplane_checkbox.isChecked()

        if not charuco_board_name:
            charuco_board_name = self._board_dropdown.currentText()

        charuco_board_definition = CHARUCO_BOARDS[charuco_board_name]()

        active_recording_info = self._get_active_recording_info()
        if active_recording_info is None:
            logger.error("Cannot calibrate from active recording - no active recording selected")
            return

        logger.info(f"Calibrating from active recording: {active_recording_info.name}")

        if not active_recording_info.synchronized_videos_status_check:
            logger.error(
                f"Cannot calibrate from {active_recording_info.name} -"
                f" `active_recording_info.synchronized_videos_status_check` is "
                f"{active_recording_info.synchronized_videos_status_check}"
            )

            return

        self._anipose_calibration_frame_worker = AniposeCalibrationThreadWorker(
            calibration_videos_folder_path=active_recording_info.synchronized_videos_folder_path,
            charuco_square_size=float(charuco_square_size_mm),
            use_charuco_as_groundplane=use_charuco_as_groundplane,
            charuco_board_definition=charuco_board_definition,
            kill_thread_event=self._kill_thread_event,
        )

        self._anipose_calibration_frame_worker.start()

        self._anipose_calibration_frame_worker.in_progress.connect(self._log_calibration_progress_callbacks)

        self._anipose_calibration_frame_worker.finished.connect(self.update_calibration_toml_path)

        self._anipose_calibration_frame_worker.groundplane_failed.connect(
            self._bring_up_groundplane_calibration_failed_dialog
        )

    def _check_active_recording_for_calibration_toml(self):
        active_rec = self._get_active_recording_info()
        if active_rec is not None:
            if active_rec.calibration_toml_check:
                return self._get_active_recording_info().calibration_toml_path


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    widget = CalibrationControlPanel()
    widget.show()
    sys.exit(app.exec())



import pytest

from freemocap.core_processes.process_motion_capture_videos.process_recording_headless import (
    process_recording_headless,
    find_calibration_toml_path,
)
from freemocap.data_layer.recording_models.recording_info_model import RecordingInfoModel
from freemocap.utilities.download_sample_data import (
    download_test_data,
)


class SessionInfo:
    sample_session_folder_path: str
    recording_info_model: RecordingInfoModel


def pytest_sessionstart():
    SessionInfo.sample_session_folder_path = download_test_data()
    calibration_toml_path = find_calibration_toml_path(SessionInfo.sample_session_folder_path)
    SessionInfo.recording_info_model = RecordingInfoModel(
        recording_folder_path=SessionInfo.sample_session_folder_path, active_tracker="mediapipe"
    )
    # TODO: make this configurable to be able to take different trackers
    process_recording_headless(
        recording_path=SessionInfo.sample_session_folder_path,
        path_to_camera_calibration_toml=calibration_toml_path,
        recording_info_model=SessionInfo.recording_info_model,
        run_blender=False,
        make_jupyter_notebook=False,
        use_tqdm=False,
    )


@pytest.fixture
def sample_session_path():
    return SessionInfo.sample_session_folder_path


@pytest.fixture
def synchronized_video_folder_path():
    return SessionInfo.recording_info_model.synchronized_videos_folder_path


@pytest.fixture
def data_folder_path():
    return SessionInfo.recording_info_model.output_data_folder_path


@pytest.fixture
def raw_skeleton_data():
    return SessionInfo.recording_info_model.raw_data_3d_npy_file_path


@pytest.fixture
def total_body_center_of_mass_data():
    return SessionInfo.recording_info_model.total_body_center_of_mass_npy_file_path


@pytest.fixture
def image_tracking_data():
    return SessionInfo.recording_info_model.data_2d_npy_file_path


@pytest.fixture
def reprojection_error_data():
    return SessionInfo.recording_info_model.reprojection_error_data_npy_file_path


if __name__ == "__main__":
    pytest_sessionstart()



import logging
import subprocess

logger = logging.getLogger(__name__)


def fix_opencv_conflict():
    logger.warning("Conflicting versions of opencv found, uninstalling `opencv-python`")
    try:
        subprocess.run(["pip", "uninstall", "-y", "opencv-python", "opencv-contrib-python"], check=True)
    except subprocess.CalledProcessError:
        logger.warning(
            "Failed to uninstall existing opencv distributions, calibration may not work without manually uninstalling them and reinstalling with `pip install opencv-contrib-python==4.8.*`"
        )
        raise

    try:
        subprocess.run(["pip", "install", "opencv-contrib-python==4.8.*"], check=True)
    except subprocess.CalledProcessError:
        logger.error(
            "Failed to install opencv-contrib-python, please run `pip install opencv-contrib-python==4.8.*` manually"
        )
        raise
    logger.info(
        "Successfully fixed opencv conflict by uninstalling all versions and reinstalling opencv-contrib-python"
    )



import time
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

import toml

from freemocap.gui.qt.utilities.save_and_load_gui_state import load_gui_state
from freemocap.system.paths_and_filenames.file_and_folder_names import (
    LOGS_INFO_AND_SETTINGS_FOLDER_NAME,
    LOG_FILE_FOLDER_NAME,
    BASE_FREEMOCAP_DATA_FOLDER_NAME,
    CALIBRATIONS_FOLDER_NAME,
    logger,
    RECORDING_SESSIONS_FOLDER_NAME,
    MOST_RECENT_RECORDING_TOML_FILENAME,
    LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME,
    OUTPUT_DATA_FOLDER_NAME,
    SYNCHRONIZED_VIDEOS_FOLDER_NAME,
    RAW_DATA_FOLDER_NAME,
    RAW_3D_NPY_FILE_NAME,
    CENTER_OF_MASS_FOLDER_NAME,
    TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME,
    DATA_2D_NPY_FILE_NAME,
    REPROJECTION_ERROR_NPY_FILE_NAME,
    DATA_3D_NPY_FILE_NAME,
    GUI_STATE_JSON_FILENAME,
)


def os_independent_home_dir():
    return str(Path.home())


def get_log_file_path():
    log_folder_path = Path(get_freemocap_data_folder_path()) / LOGS_INFO_AND_SETTINGS_FOLDER_NAME / LOG_FILE_FOLDER_NAME
    log_folder_path.mkdir(exist_ok=True, parents=True)
    log_file_path = log_folder_path / create_log_file_name()
    return str(log_file_path)


def create_log_file_name():
    return "log_" + time.strftime("%m-%d-%Y-%H_%M_%S") + ".log"


def create_camera_calibration_file_name(recording_name: str):
    return f"{recording_name}_camera_calibration.toml"


freemocap_data_folder_path = None


def get_freemocap_data_folder_path(create_folder: bool = True) -> str:
    global freemocap_data_folder_path

    try:
        if freemocap_data_folder_path is None:
            freemocap_data_folder_path = Path(os_independent_home_dir(), BASE_FREEMOCAP_DATA_FOLDER_NAME)
        if Path(get_gui_state_json_path()).exists():
            gui_state = load_gui_state(get_gui_state_json_path())
        else:  # if GUI state isn't in new location, look in old location
            gui_state = load_gui_state(
                str(Path(freemocap_data_folder_path) / LOGS_INFO_AND_SETTINGS_FOLDER_NAME / GUI_STATE_JSON_FILENAME)
            )
        freemocap_data_folder_path = Path(gui_state.freemocap_data_folder_path)
    except Exception as e:  # TODO: figure out specific exceptions to catch
        print(e)  # Cannot log this due to circular import from logging config
    finally:
        if freemocap_data_folder_path is None:
            freemocap_data_folder_path = Path(os_independent_home_dir(), BASE_FREEMOCAP_DATA_FOLDER_NAME)

    if create_folder:
        freemocap_data_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(freemocap_data_folder_path)


def get_calibrations_folder_path(create_folder: bool = True):
    calibration_folder_path = Path(get_freemocap_data_folder_path()) / CALIBRATIONS_FOLDER_NAME

    if create_folder:
        calibration_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(calibration_folder_path)


def create_new_recording_folder_path(recording_name: str):
    recording_folder_path = Path(create_new_session_folder()) / recording_name

    return str(recording_folder_path)


def get_gmt_offset_string():
    # from - https://stackoverflow.com/a/53860920/14662833
    gmt_offset_int = int(time.localtime().tm_gmtoff / 60 / 60)
    return f"{gmt_offset_int:+}"


def get_iso6201_time_string(timespec: str = "milliseconds", make_filename_friendly: bool = True):
    iso6201_timestamp = datetime.now().isoformat(timespec=timespec)
    gmt_offset_string = f"_gmt{get_gmt_offset_string()}"
    iso6201_timestamp_w_gmt = iso6201_timestamp + gmt_offset_string
    if make_filename_friendly:
        iso6201_timestamp_w_gmt = iso6201_timestamp_w_gmt.replace(":", "_")
        iso6201_timestamp_w_gmt = iso6201_timestamp_w_gmt.replace(".", "ms")
    return iso6201_timestamp_w_gmt


def create_new_default_recording_name():
    full_time = get_iso6201_time_string(timespec="seconds")
    just_hours_minutes_seconds = full_time.split("T")[1]
    recording_name = "recording_" + just_hours_minutes_seconds
    return recording_name


def session_time_tag_format():
    return "%Y-%m-%d_%H_%M_%S"


def default_session_name(string_tag: Optional[str] = None):
    if string_tag is not None:
        string_tag = f"_{string_tag}"
    else:
        string_tag = ""

    session_name = time.strftime(f"session_{session_time_tag_format()}" + string_tag)
    logger.debug(f"Creating default session name: {session_name}")

    return session_name


session_folder_path: Optional[Path] = None


def create_new_session_folder():
    global session_folder_path
    if session_folder_path is None:
        session_folder_path = Path(get_recording_session_folder_path()) / default_session_name()
    elif not session_folder_path.is_relative_to(get_freemocap_data_folder_path()):
        session_folder_path = Path(get_recording_session_folder_path()) / default_session_name()

    return str(session_folder_path)


def get_recording_session_folder_path(create_folder: bool = True):
    recording_session_folder_path = Path(get_freemocap_data_folder_path()) / RECORDING_SESSIONS_FOLDER_NAME

    if create_folder:
        recording_session_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(recording_session_folder_path)


def get_logs_info_and_settings_folder_path(create_folder: bool = True):
    logs_info_and_settings_folder_path = Path(get_freemocap_data_folder_path()) / LOGS_INFO_AND_SETTINGS_FOLDER_NAME

    if create_folder:
        logs_info_and_settings_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(logs_info_and_settings_folder_path)


def get_css_stylesheet_path():
    # return str(Path(__file__).parent.parent / "gui" / "qt" / "style_sheet" / "ElegantDark.qss")
    # return str(Path(__file__).parent.parent / "gui" / "qt" / "style_sheet" / "FunkyTown.qss")
    return str(Path(__file__).parent.parent.parent / "gui" / "qt" / "style_sheet" / "qt_style_sheet.css")


def get_scss_stylesheet_path():
    return str(Path(__file__).parent.parent.parent / "gui" / "qt" / "style_sheet" / "qt_style_sheet.scss")


def get_most_recent_recording_toml_path():
    return str(Path(get_logs_info_and_settings_folder_path()) / MOST_RECENT_RECORDING_TOML_FILENAME)


def get_gui_state_json_path():
    return str(Path(__file__).parent.parent / GUI_STATE_JSON_FILENAME)


def get_most_recent_recording_path(subfolder_str: Optional[str] = None) -> Optional[str]:
    if not Path(get_most_recent_recording_toml_path()).exists():
        logger.error(f"{MOST_RECENT_RECORDING_TOML_FILENAME} not found at {get_most_recent_recording_toml_path()}!!")
        return None

    try:
        most_recent_recording_dict = toml.load(str(get_most_recent_recording_toml_path()))
        most_recent_recording_path = most_recent_recording_dict["most_recent_recording_path"]
    except (ValueError, toml.TomlDecodeError, KeyError, FileNotFoundError) as e:
        logger.exception(e)
        return None

    if not Path(most_recent_recording_path).exists():
        logger.error(f"Most recent recording path {most_recent_recording_path} not found!!")
        return None

    if subfolder_str is not None:
        try:
            # check if folder is specified in toml file, otherwise return default
            sub_folder_path = most_recent_recording_dict[subfolder_str]
        except KeyError:
            sub_folder_path = Path(most_recent_recording_path) / subfolder_str
        return str(sub_folder_path)

    else:
        return str(most_recent_recording_path)


def get_last_successful_calibration_toml_path():
    return str(Path(get_logs_info_and_settings_folder_path()) / LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME)


def get_last_successful_calibration_name():
    path = get_last_successful_calibration_toml_path()
    if not Path(path).exists():
        logger.error(f"{LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME} not found at {path}!!")
        return None
    # toml_dict = toml.load(path)

    # TODO: this function appears unfinished, and is currently unused. Either we need to parse the TOML or remove this.


def get_blender_file_name(recording_name: str):
    return f"{recording_name}.blend"


def get_blender_file_path(recording_folder_path: Union[Path, str]):
    return str(Path(recording_folder_path) / get_blender_file_name(recording_name=Path(recording_folder_path).name))


def get_output_data_folder_path(recording_folder_path: Union[str, Path]) -> str:
    for subfolder_path in Path(recording_folder_path).iterdir():
        if subfolder_path.name == OUTPUT_DATA_FOLDER_NAME:
            return str(subfolder_path)

    raise Exception(f"Could not find a data folder in path {str(recording_folder_path)}")


def get_synchronized_videos_folder_path(recording_folder_path: Union[str, Path]) -> str:
    for subfolder_path in Path(recording_folder_path).iterdir():
        if subfolder_path.name == SYNCHRONIZED_VIDEOS_FOLDER_NAME:
            return str(subfolder_path)

    raise Exception(f"Could not find a videos folder in path {str(recording_folder_path)}")


def get_timestamps_directory(recording_directory: Union[str, Path]) -> Optional[str]:
    synchronized_videos_path = get_synchronized_videos_folder_path(recording_folder_path=recording_directory)
    if "timestamps" in [path.name for path in Path(synchronized_videos_path).iterdir()]:
        return str(Path(synchronized_videos_path) / "timestamps")
    logger.warning(f"Could not find timestamps directory in {synchronized_videos_path}")


# TODO: figure out how to deal with different file prefixes for all of these below
def get_raw_skeleton_npy_file_name(data_folder_name: Union[str, Path]) -> str:
    raw_data_subfolder_path = Path(data_folder_name) / RAW_DATA_FOLDER_NAME
    if raw_data_subfolder_path.exists:
        raw_data_npy_path_list = [path.name for path in raw_data_subfolder_path.glob("*.npy")]
        if RAW_3D_NPY_FILE_NAME in raw_data_npy_path_list:
            return str(raw_data_subfolder_path / RAW_3D_NPY_FILE_NAME)

    logger.debug(list(Path(data_folder_name).glob("*")))
    logger.debug(list(Path(raw_data_subfolder_path).glob("*")))
    raise Exception(f"Could not find a skeleton NPY file in path {str(data_folder_name)}")


def get_full_npy_file_path(output_data_folder: Union[str, Path]) -> str:
    path = Path(output_data_folder) / DATA_3D_NPY_FILE_NAME
    return str(path)


def get_total_body_center_of_mass_file_path(output_data_folder: Union[str, Path]) -> str:
    center_of_mass_subfolder_path = Path(output_data_folder) / CENTER_OF_MASS_FOLDER_NAME
    if center_of_mass_subfolder_path.exists:
        center_of_mass_npy_path_list = [path.name for path in center_of_mass_subfolder_path.glob("*.npy")]

        if TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME in center_of_mass_npy_path_list:
            return str(center_of_mass_subfolder_path / TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME)

    raise Exception(f"Could not find a total body center of mass npy file in path {str(output_data_folder)}")


def get_image_tracking_data_file_name(data_folder_name: Union[str, Path]) -> str:
    raw_data_subfolder_path = Path(data_folder_name) / RAW_DATA_FOLDER_NAME
    if raw_data_subfolder_path.exists:
        raw_data_npy_path_list = [path.name for path in raw_data_subfolder_path.glob("*.npy")]

        if DATA_2D_NPY_FILE_NAME in raw_data_npy_path_list:
            return str(raw_data_subfolder_path / DATA_2D_NPY_FILE_NAME)

    raise Exception(f"Could not find a 2d data file in path {str(data_folder_name)}")


def get_reprojection_error_file_path(data_folder_name: Union[str, Path]) -> str:
    raw_data_subfolder_path = Path(data_folder_name) / RAW_DATA_FOLDER_NAME
    if raw_data_subfolder_path.exists:
        raw_data_npy_path_list = [path.name for path in raw_data_subfolder_path.glob("*.npy")]

        if REPROJECTION_ERROR_NPY_FILE_NAME in raw_data_npy_path_list:
            return str(raw_data_subfolder_path / REPROJECTION_ERROR_NPY_FILE_NAME)

    raise Exception(f"Could not find reprojection error data file in path {str(data_folder_name)}")



from typing import Dict

from pydantic import BaseModel, model_validator

from freemocap.data_layer.skeleton_models.marker_info import MarkerInfo


class Segment(BaseModel):
    proximal: str
    distal: str


class SegmentAnthropometry(BaseModel):
    segment_com_length: float
    segment_com_percentage: float


class Segments(BaseModel):
    markers: MarkerInfo
    segment_connections: Dict[str, Segment]

    @model_validator(mode="before")
    def check_that_all_markers_exist(cls, values):
        markers = values.get("markers").all_markers
        segment_connections = values.get("segment_connections")

        for segment_name, segment_connection in segment_connections.items():
            if segment_connection.get("proximal") not in markers:
                raise ValueError(
                    f"The proximal marker {segment_connection.proximal} for {segment_name} is not in the list of markers or virtual markers."
                )
            if segment_connection.get("distal") not in markers:
                raise ValueError(
                    f"The distal marker {segment_connection.distal} for {segment_name} is not in the list of markers or virtual markers."
                )

        return values



from PySide6.QtGui import QAction

from freemocap.utilities.download_sample_data import DATASETS


class DatasetDownloadAction(QAction):
    """
    Action with built-in trigger to download a specific dataset.
    """

    def __init__(self, menu_label: str, dataset_name: str, freemocap_main_window):
        super().__init__(menu_label, freemocap_main_window)
        self.triggered.connect(lambda: freemocap_main_window.download_data(dataset_name))


def make_download_data_action(
        dataset_key: str,
        freemocap_main_window,
) -> DatasetDownloadAction:
    """
    Makes a DatasetDownloadAction for the specified dataset key.
    Raises ValueError if the dataset key is not recognized.
    """
    if dataset_key not in DATASETS:
        raise ValueError(f"Unknown dataset '{dataset_key}'. Options: {list(DATASETS)}")

    dataset = DATASETS[dataset_key]

    return DatasetDownloadAction(
        menu_label=dataset.menu_label, dataset_name=dataset_key, freemocap_main_window=freemocap_main_window
    )



import io
import logging
import zipfile
from pathlib import Path

import requests
from pydantic import BaseModel

from freemocap.system.paths_and_filenames.path_getters import get_recording_session_folder_path

logger = logging.getLogger(__name__)

TEST_DATA_NAME = "test"
DOWNLOAD_SAMPLE_DATA_ACTION_NAME = "Download Sample Data (3 cameras, ~1000 frames)"

SAMPLE_DATA_NAME = "sample"
DOWNLOAD_TEST_DATA_ACTION_NAME = "Download Test Data (3 cameras, ~200 frames)"


class Dataset(BaseModel):
    url: str
    menu_label: str
    tooltip: str = (
        "Download this dataset to use in Freemocap. The sample data contains 3 cameras and ~1000 frames. The test data contains 3 cameras and ~200 frames."
    )


DATASETS = {
    SAMPLE_DATA_NAME: Dataset(
        url="https://github.com/freemocap/skellysamples/releases/download/sample_data_v06_12_25/freemocap_sample_data.zip",
        menu_label=DOWNLOAD_SAMPLE_DATA_ACTION_NAME,
    ),
    TEST_DATA_NAME: Dataset(
        url="https://github.com/freemocap/skellysamples/releases/download/test_data_v06_09_25/freemocap_test_data.zip",
        menu_label=DOWNLOAD_TEST_DATA_ACTION_NAME,
    ),
}


def download_and_extract_zip(zip_file_url) -> str:
    try:
        logger.info(f"Downloading data from {zip_file_url}...")
        recording_session_folder_path = Path(get_recording_session_folder_path())
        recording_session_folder_path.mkdir(parents=True, exist_ok=True)
        r = requests.get(zip_file_url, stream=True, timeout=(5, 60))
        r.raise_for_status()

        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(recording_session_folder_path)

        recording_name = {
            Path(p).parts[0] for p in z.namelist() if not p.endswith("/")
        }  # gets name of recording (top-level folder in zip file)
        if len(recording_name) != 1:
            raise ValueError(f"{zip_file_url!r} contained {len(recording_name)} top-level entries: {recording_name}")

        data_path = recording_session_folder_path / recording_name.pop()
        logger.info("Data extracted successfully")
        return str(data_path)

    except requests.exceptions.RequestException as e:
        logger.error(f"Request failed: {e}")
        raise e
    except zipfile.BadZipFile as e:
        logger.error(f"Failed to unzip the file: {e}")
        raise e


def download_dataset(key: str) -> str:
    """
    Downloads the specified dataset zip file and extracts it to the recording session folder.
    Returns the path to the extracted data.
    """
    if key not in DATASETS:
        raise ValueError(f"Unknown dataset '{key}'. Options: {list(DATASETS)}")

    return download_and_extract_zip(DATASETS[key].url)


def download_test_data() -> str:
    """
    Downloads the test data zip file and extracts it to the recording session folder.
    Returns the path to the extracted data.
    """
    return download_dataset(TEST_DATA_NAME)


def download_sample_data() -> str:
    """
    Downloads the sample data zip file and extracts it to the recording session folder.
    Returns the path to the extracted data.
    """
    return download_dataset(SAMPLE_DATA_NAME)


def get_sample_data_path() -> Path:
    """
    Returns the path to the sample data folder.
    If the folder does not exist, it downloads the sample data.
    """
    sample_data_path = Path(get_recording_session_folder_path()) / "freemocap_sample_data"
    if not sample_data_path.exists():
        logger.info(f"Sample data not found at {sample_data_path}. Downloading...")
        download_sample_data()

    return sample_data_path


if __name__ == "__main__":
    sample_data_path = download_sample_data()
    print(f"Sample data downloaded successfully to path: {str(sample_data_path)}")



import logging
import subprocess
import sys
import threading

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QDialog, QPushButton, QLabel, QHBoxLayout, QFrame

from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState
from freemocap.utilities.fix_opencv_conflict import fix_opencv_conflict

logger = logging.getLogger(__name__)


class OpencvConflictDialog(QDialog):
    def __init__(self, gui_state: GuiState, kill_thread_event: threading.Event, parent=None) -> None:
        super().__init__(parent=parent)
        self.gui_state = gui_state
        self.kill_thread_event = kill_thread_event

        self.setMinimumSize(600, 300)

        self.setWindowTitle("Conflicting OpenCV Versions Found")

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        version_conflict_message = (
            "Opencv-python and opencv-contrib-python are both installed in your current environment, "
            "which will prevent you from being able to calibrate multiple cameras.\n\n"
            "Choosing `Fix OpenCV version conflict` will remove the existing OpenCV versions and install opencv-contrib-python. "
            "It will also close Freemocap to make the changes take affect, but everything will work when restarted. \n\n"
            "If you choose not to fix the conflict, you will need to manually install opencv-contrib-python and restart Freemocap "
            "in order for multi-camera calibration to work.\n\n To manually fix the conflict, close Freemocap and run `pip uninstall opencv-python opencv-contrib-python` "
            "followed by `pip install opencv-contrib-python==4.8.1.78`. Then restart Freemocap.\n\n"
        )
        welcome_text_label = QLabel(version_conflict_message)
        welcome_text_label.setWordWrap(True)
        welcome_text_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._layout.addWidget(welcome_text_label, 1)

        pypi_fix_link_string = '\n\n&#10132; <a href="https://github.com/pypa/pip/pull/10837" style="color: #333333;">A pull request to the `pypi` repo that would resolve this issue, but they won\'t do it :,(</a>'
        pypi_doc_link = QLabel(pypi_fix_link_string)
        pypi_doc_link.setOpenExternalLinks(True)
        self._layout.addWidget(pypi_doc_link)

        frame = QFrame()
        frame.setFrameShape(QFrame.Shape.HLine)
        frame.setFrameShadow(QFrame.Shadow.Sunken)
        self._layout.addWidget(frame)

        this_fix_will_label = QLabel(
            "THIS WILL PROCESS WILL: \n\n 1) UPDATE YOUR PYTHON ENVIRONMENT AND THEN \n 2)  SHUTDOWN THE FREEMOCAP GUI\n\n"
            "THE NEXT TIME YOU RUN `FREEMOCAP`, THE PROBLEM WILL BE FIXED! :D"
        )
        this_fix_will_label.setWordWrap(True)
        this_fix_will_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._layout.addWidget(this_fix_will_label)

        frame1 = QFrame()
        frame1.setFrameShape(QFrame.Shape.HLine)
        frame1.setFrameShadow(QFrame.Shadow.Sunken)
        self._layout.addWidget(frame1)

        button_box = QHBoxLayout()

        fix_opencv_button = QPushButton("Fix OpenCV conflict (Recommended)")
        fix_opencv_button.clicked.connect(self.fix_conflict_and_close)
        button_box.addWidget(fix_opencv_button)

        continue_unchanged_button = QPushButton("Continue without fixing conflict")
        continue_unchanged_button.setObjectName("continue_unchanged_button")
        continue_unchanged_button.clicked.connect(self.accept)
        button_box.addWidget(continue_unchanged_button)

        self._layout.addLayout(button_box)

    def fix_conflict_and_close(self):
        try:
            fix_opencv_conflict()
            logger.info("Successfully fixed opencv conflict, closing GUI")

        except subprocess.CalledProcessError:
            logger.error(
                "Failed to fix opencv conflict, please uninstall all opencv versions and run `pip install opencv-contrib-python==4.8.*` manually"
            )
            sys.exit(1)
        finally:
            sys.exit(0)



import logging
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
from skellytracker.trackers.base_tracker.model_info import ModelInfo

from freemocap.system.paths_and_filenames.file_and_folder_names import (
    BODY_3D_DATAFRAME_CSV_FILE_NAME,
    FACE_3D_DATAFRAME_CSV_FILE_NAME,
    LEFT_HAND_3D_DATAFRAME_CSV_FILE_NAME,
    RIGHT_HAND_3D_DATAFRAME_CSV_FILE_NAME,
)

logger = logging.getLogger(__name__)

# TODO: This is dependent on the ordering we currently use for mediapipe, maybe it should live in model_info
category_info = {
    "body": BODY_3D_DATAFRAME_CSV_FILE_NAME,
    "right_hand": RIGHT_HAND_3D_DATAFRAME_CSV_FILE_NAME,
    "left_hand": LEFT_HAND_3D_DATAFRAME_CSV_FILE_NAME,
    "face": FACE_3D_DATAFRAME_CSV_FILE_NAME,
}

DEFAULT_CATEGORY_NAME = "body"


# TODO: come up with a sensible way of splitting this for different types of trackers


def split_data(
        skeleton_3d_data: np.ndarray,
        model_info: ModelInfo,
) -> Dict[str, np.ndarray]:
    """
    Splits the skeleton data into separate arrays for each data category using the provided model info.
    Examples categories are body, left_hand, right_hand, and face.

    Parameters:
    - skeleton: The skeleton data to be split.
    - model_info: The model info used in tracking.

    Returns:
    - A dictionary where each key is a category and the value is the corresponding data array.
    """

    split_data = {}

    # TODO: this function probably shouldn't exist, and its existence implies we should be separately tracking body, hands, and face

    current_index = 0
    for category in category_info.keys():
        tracked_points_name = "num_tracked_points_" + category
        if hasattr(model_info, tracked_points_name):
            prev_index = current_index
            current_index += getattr(model_info, tracked_points_name)
            split_data[category] = skeleton_3d_data[:, prev_index:current_index, :]

    if split_data == {}:
        logger.debug("No data categories found in model info, defaulting to single 'body' category")
        split_data[DEFAULT_CATEGORY_NAME] = skeleton_3d_data

    return split_data


def create_column_names(model_info: ModelInfo) -> Dict[str, List[str]]:
    """
    Creates a list of column names for the skeleton data based on the model info.

    Parameters:
    - model_info: The model info used in tracking.

    Returns:
    - A list of column names for the skeleton data.
    """

    column_names = {}
    for category in category_info.keys():
        tracked_points_name = "num_tracked_points_" + category
        landmark_names = category + "_landmark_names"
        if hasattr(model_info, tracked_points_name):
            category_column_names = []
            if hasattr(model_info, landmark_names) and len(getattr(model_info, landmark_names)) == getattr(
                    model_info, tracked_points_name
            ):
                for name in getattr(model_info, landmark_names):
                    category_column_names.append(f"{category}_{name}_x")
                    category_column_names.append(f"{category}_{name}_y")
                    category_column_names.append(f"{category}_{name}_z")
            else:
                for i in range(getattr(model_info, tracked_points_name)):
                    category_column_names.append(f"{category}_{str(i).zfill(4)}_x")
                    category_column_names.append(f"{category}_{str(i).zfill(4)}_y")
                    category_column_names.append(f"{category}_{str(i).zfill(4)}_z")

            column_names[category] = category_column_names

    # cover the case where no data categories are found
    if column_names == {}:
        landmark_names = "landmark_names"
        tracked_points_name = "num_tracked_points"
        category_column_names = []
        if (
                hasattr(model_info, landmark_names)
                and hasattr(model_info, tracked_points_name)
                and len(getattr(model_info, landmark_names)) == getattr(model_info, tracked_points_name)
        ):
            for name in getattr(model_info, landmark_names):
                category_column_names.append(f"{category}_{name}_x")
                category_column_names.append(f"{category}_{name}_y")
                category_column_names.append(f"{category}_{name}_z")
        else:
            for i in range(getattr(model_info, tracked_points_name)):
                category_column_names.append(f"{category}_{str(i).zfill(4)}_x")
                category_column_names.append(f"{category}_{str(i).zfill(4)}_y")
                category_column_names.append(f"{category}_{str(i).zfill(4)}_z")

        column_names[DEFAULT_CATEGORY_NAME] = category_column_names

    return column_names


def save_split_csv(
        output_data_folder_path: str,
        split_data: Dict[str, np.ndarray],
        column_names: Dict[str, List[str]],
        model_info: ModelInfo,
) -> None:
    """
    Saves the split skeleton data to CSV files using column names.

    Parameters:
    - split_data: A dictionary where each key is a category and the value is the corresponding data array.
    - column_names: A dictionary where each key is a category and the value is a list of column names for the data.
    """
    if split_data == {}:
        logger.debug("No data categories found in model info, skipping saving split csv files")
        return

    number_of_frames = next(iter(split_data.values())).shape[0]

    file_prefix = model_info.name + "_"

    for category, data in split_data.items():
        column_name_list = column_names[category]
        if len(split_data) == 1:
            tracked_points_name = "num_tracked_points"
        else:
            tracked_points_name = "num_tracked_points_" + category

        data_flat = data.reshape(number_of_frames, getattr(model_info, tracked_points_name) * 3)
        dataframe = pd.DataFrame(data_flat, columns=column_name_list)
        dataframe.to_csv(str(Path(output_data_folder_path) / (file_prefix + category_info[category])), index=False)

    logger.info(f"Saved split skeleton data csvs to {output_data_folder_path}")


def save_split_npy(output_data_folder_path: str, split_data: Dict[str, np.ndarray], file_prefix: str = "") -> None:
    """
    Saves the split skeleton data to numpy files.

    Parameters:
    - split_data: A dictionary where each key is a category and the value is the corresponding data array.
    """

    if split_data == {}:
        logger.debug("No data categories found in model info, skipping saving split npy files")
        return

    if file_prefix != "" and file_prefix[-1] != "_":
        file_prefix += "_"

    for category, data in split_data.items():
        np.save(str(Path(output_data_folder_path) / (file_prefix + category + "_3d_xyz.npy")), data)

    logger.info(f"Saved split skeleton data npys to {output_data_folder_path}")


def split_and_save(
        skeleton_3d_data: np.ndarray,
        model_info: ModelInfo,
        output_data_folder_path: str,
) -> None:
    """
    Splits the skeleton data into separate arrays for each data category using the provided model info.
    Examples categories are body, left_hand, right_hand, and face.
    Then saves the split skeleton data to CSV files using column names.
    Then saves the split skeleton data to numpy files.

    Parameters:
    - skeleton: The skeleton data to be split.
    - model_info: The model info used in tracking.
    - output_data_folder_path: The path to save the split skeleton data to.
    """

    split_data_dict = split_data(skeleton_3d_data=skeleton_3d_data, model_info=model_info)
    column_names = create_column_names(model_info=model_info)
    save_split_csv(
        output_data_folder_path=output_data_folder_path,
        split_data=split_data_dict,
        column_names=column_names,
        model_info=model_info,
    )
    save_split_npy(
        output_data_folder_path=output_data_folder_path, split_data=split_data_dict, file_prefix=model_info.name
    )



import logging
import threading
from pathlib import Path

from PySide6.QtCore import Signal, QThread

from freemocap.core_processes.export_data.blender_stuff.export_to_blender.export_to_blender import export_to_blender

logger = logging.getLogger(__name__)


class ExportToBlenderThreadWorker(QThread):
    success = Signal(bool)
    in_progress = Signal(str)

    def __init__(
            self,
            recording_path: Path,
            blender_file_path: Path,
            blender_executable_path: Path,
            kill_thread_event: threading.Event,
    ):
        super().__init__()
        logger.debug("Initializing Export to Blender Thread Worker")
        self._kill_thread_event = kill_thread_event
        self.recording_path = recording_path
        self.blender_file_path = blender_file_path
        self.blender_executable_path = blender_executable_path

        self._work_done = False

    @property
    def work_done(self):
        return self._work_done

    def _emit_in_progress_data(self, message: str):
        self.in_progress.emit(message)

    def run(self):
        logger.info("Beginning to export to Blender")

        try:
            export_to_blender(
                recording_folder_path=self.recording_path,
                blender_file_path=self.blender_file_path,
                blender_exe_path=self.blender_executable_path,
            )
            self.success.emit(True)
            logger.debug("Blender Export Complete")
        except Exception as e:
            logger.exception("something went wrong in the Blender export")
            logger.error(e)
            self.success.emit(False)

        self._work_done = True



import logging
from pathlib import Path
from typing import Union, Dict, Any

import numpy as np
import pandas as pd
from skellytracker.trackers.base_tracker.model_info import ModelInfo
from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
    MediapipeModelInfo,
)

from freemocap.data_layer.data_saver.data_models import FrameData, Timestamps, Point, SkeletonSchema, ReprojectionError
from freemocap.system.paths_and_filenames.file_and_folder_names import (
    BODY_3D_DATAFRAME_CSV_FILE_NAME,
    CENTER_OF_MASS_FOLDER_NAME,
    DATA_3D_NPY_FILE_NAME,
    RIGHT_HAND_3D_DATAFRAME_CSV_FILE_NAME,
    LEFT_HAND_3D_DATAFRAME_CSV_FILE_NAME,
    FACE_3D_DATAFRAME_CSV_FILE_NAME,
    SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME,
    TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME,
    RAW_DATA_FOLDER_NAME,
    REPROJECTION_ERROR_NPY_FILE_NAME,
    FULL_REPROJECTION_ERROR_NPY_FILE_NAME,
    REPROJECTION_FILTERED_PREFIX,
)
from freemocap.system.paths_and_filenames.path_getters import (
    get_output_data_folder_path,
    get_timestamps_directory,
)

logger = logging.getLogger(__name__)

# TODO: Need to generalize this beyond mediapipe, and make COM data optional
mediapipe_model_info = MediapipeModelInfo()


class DataLoader:
    def __init__(
        self,
        recording_folder_path: Union[str, Path],
        include_hands: bool = True,
        include_face: bool = True,
        include_com: bool = True,
        include_reprojection_error: bool = True,
        model_info: ModelInfo = mediapipe_model_info,
    ):
        self._recording_folder_path = Path(recording_folder_path)
        self.include_hands = include_hands
        self.include_face = include_face
        self.include_com = include_com
        self.include_reprojection_error = include_reprojection_error
        self._model_info = model_info

        self._recording_name = self._recording_folder_path.name
        self._output_folder_path = Path(get_output_data_folder_path(self._recording_folder_path))
        self._raw_data_folder_path = self._output_folder_path / RAW_DATA_FOLDER_NAME
        self.number_of_frames = None

        self._set_file_prefix()
        self._load_data()
        self._set_reprojection_error_point_names()

    def _load_data(self):
        self._load_timestamps()
        self._load_data_frames()
        self._load_full_npy_data()
        self._load_center_of_mass_data()
        self._load_reprojection_error_data()
        # self._load_segment_lengths()
        # self._load_skeleton_schema()
        self._validate_data()

    def _load_data_frames(self):
        self.body_dataframe = self._load_dataframe(self._file_prefix + BODY_3D_DATAFRAME_CSV_FILE_NAME)
        self.number_of_frames = len(self.body_dataframe)
        if self.include_hands:
            try:
                self.right_hand_dataframe = self._load_dataframe(
                    self._file_prefix + RIGHT_HAND_3D_DATAFRAME_CSV_FILE_NAME
                )
                self.left_hand_dataframe = self._load_dataframe(
                    self._file_prefix + LEFT_HAND_3D_DATAFRAME_CSV_FILE_NAME
                )
            except FileNotFoundError:
                logger.warning("Unable to load hand data from file.")
                self.right_hand_dataframe = None
                self.left_hand_dataframe = None
                self.include_hands = False
        if self.include_face:
            try:
                self.face_dataframe = self._load_dataframe(self._file_prefix + FACE_3D_DATAFRAME_CSV_FILE_NAME)
            except FileNotFoundError:
                logger.warning("Unable to load face data from file.")
                self.face_dataframe = None
                self.include_face = False

    def _load_dataframe(self, filename):
        return pd.read_csv(self._output_folder_path / filename)

    def _validate_data(self):
        self._validate_dataframe(self.right_hand_dataframe, "right hand")
        self._validate_dataframe(self.left_hand_dataframe, "left hand")
        self._validate_dataframe(self.face_dataframe, "face")
        self._validate_numpy_array(self.center_of_mass_xyz, "center of mass")
        self._validate_numpy_array(self.segment_center_of_mass_segment_xyz, "segment center of mass")
        self._validate_numpy_array(self.reprojection_error_frame_name_value, "average reprojection error")
        self._validate_numpy_array(
            self.reprojection_error_camera_frame_name_value, "reprojection error by camera", frame_index=1
        )

    def _validate_dataframe(self, df, df_name):
        if df is not None and len(df) != self.number_of_frames:
            raise ValueError(
                f"The number of frames in the {df_name} dataframe ({len(df)}) is different from the expected number of frames ({self.number_of_frames})."
            )

    def _validate_numpy_array(self, np_array, np_array_name, frame_index: int = 0):
        if np_array is not None and np_array.shape[frame_index] != self.number_of_frames:
            print(f"\n\n {np_array.shape} \n\n")
            raise ValueError(
                f"The number of frames in the {np_array_name} data ({np_array.shape[frame_index]}) is different from the expected number of frames ({self.number_of_frames})."
            )

    def _load_timestamps(self):
        timestamps_directory = get_timestamps_directory(recording_directory=self._recording_folder_path)
        if timestamps_directory is None:
            logger.warning("No timestamps directory found. Skipping timestamps loading.")
            return

        timestamps_by_camera = {}
        for timestamps_npy in Path(timestamps_directory).glob("*.npy"):
            camera_name = timestamps_npy.stem
            timestamps_by_camera[camera_name] = np.load(str(timestamps_npy))

        self._timestamps_by_camera = timestamps_by_camera
        timestamps = [timestamps for timestamps in timestamps_by_camera.values()]
        self._timestamps_mean_per_frame = np.nanmean(np.asarray(timestamps), axis=0)

    def _load_center_of_mass_data(self):
        """
        Load additional data like center of mass and segment lengths.
        """
        try:
            self.center_of_mass_xyz = np.load(
                self._output_folder_path
                / CENTER_OF_MASS_FOLDER_NAME
                / (self._file_prefix + TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME)
            )
            self.segment_center_of_mass_segment_xyz = np.load(
                self._output_folder_path
                / CENTER_OF_MASS_FOLDER_NAME
                / (self._file_prefix + SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME)
            )
        except FileNotFoundError:
            logger.warning("Unable to load center of mass data from file.")
            self.center_of_mass_xyz = None
            self.segment_center_of_mass_segment_xyz = None
            self.include_com = False

    def _load_reprojection_error_data(self):
        """
        Load reprojection error averaged across cameras and by camera.
        Defaults to using reprojection filtered values if they exist.
        """
        filtered_average_error_path = (
            self._raw_data_folder_path
            / f"{REPROJECTION_FILTERED_PREFIX}{self._file_prefix}{REPROJECTION_ERROR_NPY_FILE_NAME}"
        )
        filtered_by_camera_path = (
            self._raw_data_folder_path
            / f"{REPROJECTION_FILTERED_PREFIX}{self._file_prefix}{FULL_REPROJECTION_ERROR_NPY_FILE_NAME}"
        )

        raw_average_error_path = self._raw_data_folder_path / f"{self._file_prefix}{REPROJECTION_ERROR_NPY_FILE_NAME}"
        raw_by_camera_path = self._raw_data_folder_path / f"{self._file_prefix}{FULL_REPROJECTION_ERROR_NPY_FILE_NAME}"

        try:
            if filtered_average_error_path.exists():
                logger.debug(f"Loading filtered average reprojection error from {filtered_average_error_path}")
                self.reprojection_error_frame_name_value = np.load(filtered_average_error_path)
            else:
                logger.debug(f"Loading unfiltered average reprojection error from {raw_average_error_path}")
                self.reprojection_error_frame_name_value = np.load(raw_average_error_path)
            if filtered_by_camera_path.exists():
                logger.debug(f"Loading filtered by camera reprojection error from {filtered_by_camera_path}")
                self.reprojection_error_camera_frame_name_value = np.load(filtered_by_camera_path)
            else:
                logger.debug(f"Loading unfiltered by camera reprojection error from {raw_by_camera_path}")
                self.reprojection_error_camera_frame_name_value = np.load(raw_by_camera_path)
        except FileNotFoundError:
            logger.warning("unable to load reprojection error data from file.")
            self.reprojection_error_frame_name_value = None
            self.reprojection_error_camera_frame_name_value = None
            self.include_reprojection_error = None

    def _load_full_npy_data(self):
        self.data_frame_name_xyz = np.load(self._output_folder_path / (self._file_prefix + DATA_3D_NPY_FILE_NAME))

        if not self.number_of_frames:
            self.number_of_frames = self.data_frame_name_xyz.shape[0]

    def load_frame_data(self, frame_number: int) -> FrameData:
        return FrameData(
            timestamps=self.get_timestamps(frame_number),
            tracked_points=self.get_tracked_points(frame_number),
            reprojection_error=self._get_reprojection_error_data(frame_number),
        )

    def get_timestamps(self, frame_number):
        try:
            return Timestamps(
                mean=self._timestamps_mean_per_frame[frame_number],
                by_camera={
                    camera_name: timestamps[frame_number]
                    for camera_name, timestamps in self._timestamps_by_camera.items()
                },
            )
        except AttributeError:
            return Timestamps(mean=None)

    def get_tracked_points(self, frame_number) -> Dict[str, Point]:
        right_hand_points = {}
        left_hand_points = {}
        face_points = {}
        center_of_mass_points = {}

        body_points = self._process_dataframe(self.body_dataframe.iloc[frame_number, :])

        if self.include_com:
            center_of_mass_points = self._get_center_of_mass_data(frame_number)

        if self.include_hands:
            left_hand_points, right_hand_points = self._load_hand_data(frame_number)

        if self.include_face:
            face_points = self._process_dataframe(self.face_dataframe.iloc[frame_number, :])

        all_points = {}
        all_points.update(body_points)
        all_points.update(center_of_mass_points)
        all_points.update(right_hand_points)
        all_points.update(left_hand_points)
        all_points.update(face_points)

        return all_points

    def _load_hand_data(self, frame_number: int):
        right_hand_points = self._process_dataframe(self.right_hand_dataframe.iloc[frame_number, :])
        left_hand_points = self._process_dataframe(self.left_hand_dataframe.iloc[frame_number, :])
        return left_hand_points, right_hand_points

    def _get_center_of_mass_data(self, frame_number: int):
        """
        Calculate the center of mass for a given frame number.
        """
        com_data = {}
        com_data["com_full_body"] = Point(
            x=self.center_of_mass_xyz[frame_number, 0],
            y=self.center_of_mass_xyz[frame_number, 1],
            z=self.center_of_mass_xyz[frame_number, 2],
        )

        if self._model_info.center_of_mass_definitions is None:
            raise ValueError("center of mass definitions are not provided.")

        for segment_number, segment_name in enumerate(self._model_info.center_of_mass_definitions.keys()):
            com_data[f"com_{segment_name}"] = Point(
                x=self.segment_center_of_mass_segment_xyz[frame_number, segment_number, 0],
                y=self.segment_center_of_mass_segment_xyz[frame_number, segment_number, 1],
                z=self.segment_center_of_mass_segment_xyz[frame_number, segment_number, 2],
            )

        return com_data

    def _get_reprojection_error_data(self, frame_number: int) -> Dict[str, ReprojectionError]:
        """
        Get reprojection error data for a given frame number.
        """
        if self.reprojection_error_frame_name_value is None:
            return {}
        reprojection_error_data = {}
        for index, point_name in enumerate(self._reprojection_error_point_names):
            reprojection_error_data[point_name] = ReprojectionError(
                value=self.reprojection_error_frame_name_value[frame_number, index]
            )
        return reprojection_error_data

    def _process_dataframe(self, data_frame) -> Dict[str, Any]:
        """
        Process a single row from a dataframe and add it to the recording_data_by_frame_number dictionary.

        Args:
            data_frame (DataFrame): the dataframe to process.
        """
        tracked_points = {}
        column_names = list(data_frame.index)
        for column_name in column_names:
            point_name = column_name[:-2]  # e.g. "nose" or "left_knee" or whatever
            dimension = column_name[-1]  # e.g. "x" or "y" or "z"

            value = data_frame[column_name]
            if pd.isna(value):  # Check if value is NaN
                value = None  # Replace NaN with None which is valid in JSON

            tracked_points.setdefault(point_name, {})[dimension] = value

        return tracked_points

    def get_data_by_frame(self):
        recording_data_by_frame_number = {}
        for frame_number in range(self.number_of_frames):
            recording_data_by_frame_number[frame_number] = self.load_frame_data(frame_number).to_dict()
        return recording_data_by_frame_number

    def _load_skeleton_schema(self):  # TODO: replace this with updated Skeleton class
        self.skeleton_schema = SkeletonSchema(schema_dict=MediapipeModelInfo.skeleton_schema)

    def _set_file_prefix(self) -> None:
        self._file_prefix = self._model_info.name

        if self._file_prefix[-1] != "_":
            self._file_prefix += "_"

    def _set_reprojection_error_point_names(self) -> None:
        landmark_names = []
        landmark_names.extend(self._model_info.body_landmark_names)

        if self.include_hands:
            right_hand_names = [f"right_{i:04d}" for i in range(21)]
            landmark_names.extend(right_hand_names)

            left_hand_names = [f"left_{i:04d}" for i in range(21)]
            landmark_names.extend(left_hand_names)

        if self.include_face:
            face_names = [f"{i:04d}" for i in range(478)]
            landmark_names.extend(face_names)

        self._reprojection_error_point_names = landmark_names



from pathlib import Path
from typing import Union

import pytest

from freemocap.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)
from freemocap.utilities.get_video_paths import get_video_paths


@pytest.mark.usefixtures("synchronized_video_folder_path")
def test_synchronized_video_frame_counts(synchronized_video_folder_path: Union[Path, str]):
    """
    Test if all the videos in this folder have precisely the same number of frames
    """
    list_of_video_paths = get_video_paths(Path(synchronized_video_folder_path))

    assert len(list_of_video_paths) > 0, f"No videos found in {synchronized_video_folder_path}"

    frame_counts = list(get_number_of_frames_of_videos_in_a_folder(synchronized_video_folder_path).values())

    assert (
            len(set(frame_counts)) == 1
    ), f"Videos in {synchronized_video_folder_path} have different frame counts: {frame_counts}"



import logging

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QWidget, QLabel, QTabWidget

# from skelly_viewer import SkellyViewer
from skellycam import SkellyCamWidget

from freemocap.gui.qt.widgets.home_widget import (
    HomeWidget,
)

logger = logging.getLogger(__name__)


class CentralTabWidget(QTabWidget):
    def __init__(
        self,
        skelly_cam_widget: SkellyCamWidget,
        camera_controller_widget: QWidget,
        welcome_to_freemocap_widget: HomeWidget,
        skelly_viewer_widget: QWidget,
        directory_view_widget: QWidget,
        active_recording_info_widget: QWidget,
        parent=None,
    ):
        super().__init__(parent=parent)
        self.parent = parent

        # self.setTabShape(QTabWidget.TabShape.Triangular)

        self._skelly_cam_widget = skelly_cam_widget
        self._camera_controller_widget = camera_controller_widget
        self._welcome_to_freemocap_widget = welcome_to_freemocap_widget
        self._skelly_viewer_widget = skelly_viewer_widget
        self._directory_view_widget = directory_view_widget
        self._active_recording_info_widget = active_recording_info_widget

        self._create_welcome_tab(self)
        self._create_skellycam_view_tab(self)
        self._create_skelly_viewer_tab(self)
        self._create_directory_view_tab(self)
        self._create_active_recording_info_tab(self)

    def set_welcome_tab_enabled(self, enabled: bool):
        self.setTabEnabled(0, enabled)

    def set_camera_view_tab_enabled(self, enabled: bool):
        self.setTabEnabled(1, enabled)

    def set_visualize_data_tab_enabled(self, enabled: bool):
        self.setTabEnabled(2, enabled)

    def _create_welcome_tab(self, tab_widget: QTabWidget):
        logger.debug("Creating welcome tab")
        tab_widget.addTab(self._welcome_to_freemocap_widget, "Home")

    def _create_skellycam_view_tab(self, tab_widget: QTabWidget):
        logger.debug("Creating skellycam view tab")
        dummy_widget = QWidget()
        self._camera_view_layout = QVBoxLayout()
        self._camera_view_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        dummy_widget.setLayout(self._camera_view_layout)
        tab_widget.addTab(dummy_widget, "Cameras")
        # tab_widget.setToolTip(skellycam.__repo_url__)

        # self._qt_multi_camera_viewer_widget.resize(1280, 720)
        self._camera_view_layout.addWidget(self._camera_controller_widget)

        self._camera_view_layout.addWidget(self._skelly_cam_widget)

        lag_note_label = QLabel(
            "NOTE: If you experience lag in your camera views, decrease the resolution and/or use fewer cameras. The frames are likely being recorded properly, its just the viewer that is lagging. A fix is incoming soon!"
        )
        lag_note_label.setStyleSheet("font-size: 10px;")
        lag_note_label.setWordWrap(True)
        layout = QVBoxLayout()
        layout.addWidget(lag_note_label)
        layout.setAlignment(Qt.AlignmentFlag.AlignBottom)
        self._camera_view_layout.addLayout(layout)
        self._camera_view_layout.setContentsMargins(20, 0, 20, 0)

    def _create_skelly_viewer_tab(self, tab_widget: QTabWidget):
        logger.debug("Creating export_data tab")
        tab_widget.addTab(self._skelly_viewer_widget, "Data Viewer")
        # tab_widget.setToolTip(skelly_viewer.__repo_url__)

    def _create_directory_view_tab(self, tab_widget: QTabWidget):
        logger.debug("Creating directory view tab")
        tab_widget.addTab(self._directory_view_widget, "Directory View")

    def _create_active_recording_info_tab(self, tab_widget: QTabWidget):
        logger.debug("Creating active recording info tab")
        tab_widget.addTab(self._active_recording_info_widget, "Active Recording Info")



from pathlib import Path
from typing import Union

import numpy as np
import pytest

from freemocap.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)


@pytest.mark.usefixtures("raw_skeleton_data", "reprojection_error_data")
def test_skeleton_data_exists(
        raw_skeleton_data: Union[str, Path],
        reprojection_error_data: Union[str, Path],
):
    """
    test that the `3d detection` process worked correctly by checking that there is an `.npy` file containing the 3d data in the `output_data_folder`
    """

    assert Path(
        raw_skeleton_data
    ).is_file(), f"3d skeleton data file does not exist at {raw_skeleton_data}"

    assert Path(
        reprojection_error_data
    ).is_file(), f"3d skeleton reprojection error data file does not exist at {reprojection_error_data}"


@pytest.mark.usefixtures("synchronized_video_folder_path", "raw_skeleton_data", "reprojection_error_data")
def test_skeleton_data_shape(
        synchronized_video_folder_path: Union[str, Path],
        raw_skeleton_data: Union[str, Path, np.ndarray],
        reprojection_error_data: Union[str, Path, np.ndarray],
):
    """
    test that the `3d detection` process worked correctly by checking:
    1. Dimensions of the raw skeleton numpy array are correct
    2. Dimensions of the reprojection error numpy array are correct
    """
    if isinstance(raw_skeleton_data, (str, Path)):
        skel3d_frame_marker_xyz = np.load(raw_skeleton_data)
    else:
        skel3d_frame_marker_xyz = raw_skeleton_data

    assert (
            len(skel3d_frame_marker_xyz.shape) == 3
    ), f"3d skeleton data file should have 3 dimensions, but has shape  {skel3d_frame_marker_xyz.shape}"

    if isinstance(reprojection_error_data, (str, Path)):
        skeleton_reprojection_error_fr_mar = np.load(reprojection_error_data)
    else:
        skeleton_reprojection_error_fr_mar = reprojection_error_data

    assert (
            len(skeleton_reprojection_error_fr_mar.shape) == 2
    ), f"3d skeleton reprojection error data file should have 2 dimensions but has shape {skeleton_reprojection_error_fr_mar.shape}"

    frame_counts = list(get_number_of_frames_of_videos_in_a_folder(synchronized_video_folder_path).values())
    assert (
            len(set(frame_counts)) == 1
    ), f"Videos in {synchronized_video_folder_path} have different frame counts: {frame_counts}"

    number_of_frames = frame_counts[0]

    assert skel3d_frame_marker_xyz.shape[0] == number_of_frames
    assert skeleton_reprojection_error_fr_mar.shape[0] == number_of_frames

    # TODO - check number of tracked points vs 'expected' number of tracked points

    assert (
            skel3d_frame_marker_xyz.shape[2] == 3
    ), f"3d skeleton data file should have 3 dimensions for X,Y,Z, but instead has {skel3d_frame_marker_xyz.shape[2]}"



import logging

import cv2
from fastapi import APIRouter
from pydantic import BaseModel
from src.api.services.user_config import UserConfigService, WebcamConfigModel
from src.cameras.detection.cam_singleton import get_or_create_cams
from src.cameras.multicam_manager.cv_camera_manager import OpenCVCameraManager

camera_router = APIRouter()

logger = logging.getLogger(__name__)


class CameraPreviewModel(BaseModel):
    session_id: str = None


@camera_router.post("/camera/config/{session_id}")
async def config_cam(webcam_config_model: WebcamConfigModel, session_id):
    s = UserConfigService()
    return s.save_webcam_config_to_disk(webcam_config_model, session_id)


@camera_router.post("/camera/show_camera/{webcam_id}")
async def show_camera(
        webcam_id: str, camera_preview_model: CameraPreviewModel = CameraPreviewModel()
):
    cv_cam_manager = OpenCVCameraManager(session_id=camera_preview_model.session_id)
    with cv_cam_manager.start_capture_session_single_cam(
            webcam_id
    ) as connected_camera_and_writer:
        should_continue = True
        this_camera = connected_camera_and_writer.cv_camera
        cv2_window_name = "(ESC to close) PREVIEWING camera_" + webcam_id
        while should_continue:
            if not this_camera.new_frame_ready:
                continue
            cv2.imshow(cv2_window_name, this_camera.latest_frame.image)
            # exit loop when user presses ESC key
            exit_key = cv2.waitKey(1)
            if exit_key == 27:
                logger.info("ESC has been pressed.")
                should_continue = False
                cv2.destroyWindow(cv2_window_name)


@camera_router.post("/camera/show_cameras")
async def cv2_imshow_all_camera(
        camera_preview_model: CameraPreviewModel = CameraPreviewModel(),
):
    cv_cam_manager = OpenCVCameraManager(session_id=camera_preview_model.session_id)
    with cv_cam_manager.start_capture_session_all_cams() as connected_cameras_dict:
        logger.info(f"Available cameras: {connected_cameras_dict}")
        should_continue = True
        while should_continue:
            for this_webcam_id, this_open_cv_camera in connected_cameras_dict.items():
                exit_key = cv2.waitKey(1)
                if exit_key == 27:
                    logger.info("ESC has been pressed.")
                    should_continue = False
                    cv2.destroyWindow(cv2_window_name)
                if not this_open_cv_camera.new_frame_ready:
                    continue
                cv2_window_name = "(ESC to close) PREVIEWING camera_" + this_webcam_id
                cv2.imshow(cv2_window_name, this_open_cv_camera.latest_frame.image)
                exit_key = cv2.waitKey(1)
                if exit_key == 27:
                    logger.info("ESC has been pressed.")
                    should_continue = False
                    cv2.destroyWindow(cv2_window_name)


@camera_router.get("/camera/detect")
async def get_cameras():
    return get_or_create_cams()


@camera_router.get("/camera/redetect")
async def redetect_cameras():
    return get_or_create_cams(always_create=True)



from pathlib import Path
from typing import Callable, Union

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.anipose_camera_calibrator import (
    AniposeCameraCalibrator,
    GroundPlaneSuccess,
)
from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)


async def async_run_anipose_capture_volume_calibration(**kwargs):
    run_anipose_capture_volume_calibration(**kwargs)


def run_anipose_capture_volume_calibration(
        charuco_board_definition: CharucoBoardDefinition,
        charuco_square_size: float,
        calibration_videos_folder_path: Union[str, Path],
        pin_camera_0_to_origin: bool = True,
        use_charuco_as_groundplane: bool = False,
        progress_callback: Callable[[str], None] = lambda _: None,
) -> tuple[Path, GroundPlaneSuccess]:
    anipose_camera_calibrator = AniposeCameraCalibrator(
        charuco_board_definition,
        charuco_square_size=charuco_square_size,
        calibration_videos_folder_path=calibration_videos_folder_path,
        progress_callback=progress_callback,
    )
    toml_path, groundplane_success = anipose_camera_calibrator.calibrate_camera_capture_volume(
        pin_camera_0_to_origin=pin_camera_0_to_origin, use_charuco_as_groundplane=use_charuco_as_groundplane
    )

    return toml_path, groundplane_success



import logging
import multiprocessing
from typing import Optional

from freemocap.core_processes.post_process_skeleton_data.post_process_skeleton import post_process_data
from freemocap.core_processes.process_motion_capture_videos.processing_pipeline_functions.anatomical_data_pipeline_functions import (
    calculate_anatomical_data,
)
from freemocap.core_processes.process_motion_capture_videos.processing_pipeline_functions.data_saving_pipeline_functions import (
    save_data,
)
from freemocap.core_processes.process_motion_capture_videos.processing_pipeline_functions.image_tracking_pipeline_functions import (
    run_image_tracking_pipeline,
)
from freemocap.core_processes.process_motion_capture_videos.processing_pipeline_functions.pipeline_check import (
    processing_pipeline_check,
)
from freemocap.core_processes.process_motion_capture_videos.processing_pipeline_functions.triangulation_pipeline_functions import (
    get_triangulated_data,
)
from freemocap.data_layer.data_saver.data_saver import DataSaver
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.system.logging.configure_logging import log_view_logging_format_string
from freemocap.system.logging.queue_logger import DirectQueueHandler
from freemocap.utilities.kill_event_exception import KillEventException

logger = logging.getLogger(__name__)


def process_recording_folder(
        recording_processing_parameter_model: ProcessingParameterModel,
        kill_event: multiprocessing.Event = None,
        logging_queue: Optional[multiprocessing.Queue] = None,
        use_tqdm: bool = True,
) -> None:
    """

    Parameters
    ----------
    recording_processing_parameter_model : ProcessingParameterModel
        ProcessingParameterModel object (contains all the paths and parameters necessary to process a session folder)

    kill_event : multiprocessing.Event
        Event to kill the process

    logging_queue : multiprocessing.Queue
        Queue to communicate logging between processes

    use_tqdm : bool
        Whether or not to use tqdm to show progress bar in terminal

    """
    if logging_queue:
        handler = DirectQueueHandler(logging_queue)
        handler.setFormatter(logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(handler)

    try:
        processing_pipeline_check(processing_parameters=recording_processing_parameter_model)
    except FileNotFoundError as e:
        logger.error(
            "processing parameters are not valid for recording status"
            f" ::: {e}")
        if logging_queue:
            logging_queue.put(e)
        raise e

    try:
        image_data_numCams_numFrames_numTrackedPts_XYZ = run_image_tracking_pipeline(
            processing_parameters=recording_processing_parameter_model,
            kill_event=kill_event,
            queue=logging_queue,
            use_tqdm=use_tqdm,
        )
    except (RuntimeError, ValueError) as e:
        logger.error("2D skeleton detection failed, cannot continue processing")
        if logging_queue:
            logging_queue.put(e)
        raise e

    if kill_event is not None and kill_event.is_set():
        exception = KillEventException("Process was killed")
        if logging_queue:
            logging_queue.put(exception)
        raise exception

    try:
        raw_skel3d_frame_marker_xyz = get_triangulated_data(
            image_data_numCams_numFrames_numTrackedPts_XYZ=image_data_numCams_numFrames_numTrackedPts_XYZ,
            processing_parameters=recording_processing_parameter_model,
            kill_event=kill_event,
            queue=logging_queue,
        )
    except (RuntimeError, FileNotFoundError, ValueError) as e:
        logger.error("Triangulation failed, cannot continue processing")
        if logging_queue:
            logging_queue.put(e)
        raise e

    if kill_event is not None and kill_event.is_set():
        exception = KillEventException("Process was killed")
        if logging_queue:
            logging_queue.put(exception)
        raise exception

    try:
        # TODO: find out if skellyforge does all the error handling we need - if not add it to post_process_data
        skel3d_frame_marker_xyz = post_process_data(
            recording_processing_parameter_model=recording_processing_parameter_model,
            raw_skel3d_frame_marker_xyz=raw_skel3d_frame_marker_xyz,
            queue=logging_queue,
        )
    except (RuntimeError, ValueError, AttributeError) as e:
        logger.error("Post processing failed, cannot continue processing")
        if logging_queue:
            logging_queue.put(e)
        raise e

    if kill_event is not None and kill_event.is_set():
        exception = KillEventException("Process was killed")
        if logging_queue:
            logging_queue.put(exception)
        raise exception

    try:
        anatomical_data_dict = calculate_anatomical_data(
            processing_parameters=recording_processing_parameter_model,
            skel3d_frame_marker_xyz=skel3d_frame_marker_xyz,
            queue=logging_queue,
        )
    except (RuntimeError, ValueError, TypeError, KeyError, AttributeError) as e:
        logger.error("Anatomical data calculation failed, cannot continue processing")
        if logging_queue:
            logging_queue.put(e)
        raise e

    if kill_event is not None and kill_event.is_set():
        exception = KillEventException("Process was killed")
        if logging_queue:
            logging_queue.put(exception)
        raise exception

    # TODO: deprecate save_data function in favor of DataSaver
    save_data(
        skel3d_frame_marker_xyz=skel3d_frame_marker_xyz,
        segment_COM_frame_imgPoint_XYZ=anatomical_data_dict["segment_COM"],
        totalBodyCOM_frame_XYZ=anatomical_data_dict["total_body_COM"],
        rigid_bones_data=anatomical_data_dict["rigid_bones_data"],
        processing_parameters=recording_processing_parameter_model,
        queue=logging_queue,
    )
    DataSaver(
        recording_folder_path=recording_processing_parameter_model.recording_info_model.path,
        model_info=recording_processing_parameter_model.tracking_model_info,
    ).save_all()

    logger.info(f"Done processing {recording_processing_parameter_model.recording_info_model.path}")



from pathlib import Path
import numpy as np
from time import perf_counter_ns

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.freemocap_anipose import CameraGroup


if __name__ == "__main__":
    # TODO: do all this setup with recording_info object
    recording_folder = Path("/Users/philipqueen/freemocap_data/recording_sessions/freemocap_test_data/")

    calibration_toml_path = list(recording_folder.glob("*calibration.toml"))[0]
    calibration = CameraGroup.load(calibration_toml_path)

    data_2d_path = (
        recording_folder
        / "output_data"
        / "raw_data"
        / "mediapipe_2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy"
    )
    data_2d = np.load(data_2d_path)
    data_2d_flat = data_2d.reshape(data_2d.shape[0], -1, 2)

    data_3d = calibration.triangulate(data_2d_flat, progress=False)
    print(f"data_3d shape: {data_3d.shape}")
    print(f"data_2d_flat shape: {data_2d_flat.shape}")

    num_trials = 5

    print("Benchmarking reprojection error with mean=True:\n")
    times = []
    for i in range(num_trials):
        start = perf_counter_ns()
        calibration.reprojection_error(points_3d=data_3d, points_2d=data_2d_flat, mean=True)
        end = perf_counter_ns()
        times.append((end - start) / 1e9)
        print(f"This round reprojection error time is: {(end - start) / 1e9} seconds for {data_2d.shape[1]} frames")
        print(f"This round average time per frame is: {(end - start) / 1e9 / data_2d.shape[1]} seconds\n")

    print(
        f"\tAverage reprojection error time across trials is: {np.mean(times)} seconds for {data_2d.shape[1]} frames over {num_trials} runs"
    )
    print(f"\tAverage time per frame is: {np.mean(times) / (data_2d.shape[1] * num_trials)} seconds\n")

    print("Benchmarking reprojection error with mean=False:\n")
    times = []
    for i in range(num_trials):
        start = perf_counter_ns()
        calibration.reprojection_error(points_3d=data_3d, points_2d=data_2d_flat, mean=False)
        end = perf_counter_ns()
        times.append((end - start) / 1e9)
        print(f"This round reprojection error time is: {(end - start) / 1e9} seconds for {data_2d.shape[1]} frames")
        print(f"This round average time per frame is: {(end - start) / 1e9 / data_2d.shape[1]} seconds\n")
    print(
        f"\tAverage reprojection error time across trials is: {np.mean(times)} seconds for {data_2d.shape[1]} frames over {num_trials} runs"
    )
    print(f"\tAverage time per frame is: {np.mean(times) / (data_2d.shape[1] * num_trials)} seconds\n")



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, model_validator, Field, field_validator


class VirtualMarkerInfo(BaseModel):
    virtual_markers: Dict[str, Dict[str, List[Union[float, str]]]]

    @field_validator("virtual_markers")
    @classmethod
    def validate_virtual_marker(cls, virtual_marker: Dict[str, List[Union[float, str]]]):
        for virtual_marker_name, virtual_marker_values in virtual_marker.items():
            marker_names = virtual_marker_values.get("marker_names", [])  # TODO: make sure this .get is valid
            marker_weights = virtual_marker_values.get("marker_weights", [])

            if len(marker_names) != len(marker_weights):
                raise ValueError(
                    f"The number of marker names must match the number of marker weights for {virtual_marker_name}. Currently there are {len(marker_names)} names and {len(marker_weights)} weights."
                )

            if not isinstance(marker_names, list) or not all(isinstance(name, str) for name in marker_names):
                raise ValueError(f"Marker names must be a list of strings for {marker_names}.")

            if not isinstance(marker_weights, list) or not all(
                    isinstance(weight, (int, float)) for weight in marker_weights
            ):
                raise ValueError(f"Marker weights must be a list of numbers for {virtual_marker_name}.")

            weight_sum = sum(marker_weights)
            if not 0.99 <= weight_sum <= 1.01:  # Allowing a tiny bit of floating-point leniency
                raise ValueError(
                    f"Marker weights must sum to approximately 1 for {virtual_marker_name} Current sum is {weight_sum}."
                )

        return virtual_marker


class MarkerInfo(BaseModel):
    original_marker_names: List[str]
    virtual_marker_definition: Optional[VirtualMarkerInfo] = None
    all_markers_list: List[str] = Field(default_factory=list)

    @model_validator(mode="before")
    def copy_markers_to_all(cls, values):
        """Copy markers to all_markers_list at initialization."""
        original_marker_names = values.get("original_marker_names")
        if original_marker_names:
            # Directly initializing all_markers_list with a copy of original_marker_names
            values["all_markers_list"] = original_marker_names.copy()
        return values

    def add_virtual_markers(self, virtual_markers_dict: Dict[str, Dict[str, List[Union[float, str]]]]):
        """Add virtual markers and update all_markers_list."""
        self.virtual_marker_definition = VirtualMarkerInfo(virtual_markers=virtual_markers_dict)
        for virtual_marker_name in self.virtual_marker_definition.virtual_markers.keys():
            if virtual_marker_name not in self.all_markers_list:
                self.all_markers_list.append(virtual_marker_name)

    @property
    def all_markers(self) -> List[str]:
        """Publicly expose the combined list of markers."""
        return self.all_markers_list

    @classmethod
    def create(cls, marker_list: List[str]):
        """Class method to create an instance with initial marker names."""
        return cls(original_marker_names=marker_list)



import logging
import multiprocessing
from pathlib import Path
from typing import Optional

import numpy as np

from freemocap.core_processes.post_process_skeleton_data.post_process_skeleton import save_numpy_array_to_disk
from freemocap.core_processes.post_process_skeleton_data.split_and_save import split_and_save
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.system.logging.configure_logging import log_view_logging_format_string
from freemocap.system.logging.queue_logger import DirectQueueHandler
from freemocap.system.paths_and_filenames.file_and_folder_names import (
    CENTER_OF_MASS_FOLDER_NAME,
    DATA_3D_NPY_FILE_NAME,
    RIGID_BONES_NPY_FILE_NAME,
    SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME,
    TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME,
)

logger = logging.getLogger(__name__)


def save_data(
        skel3d_frame_marker_xyz: np.ndarray,
        segment_COM_frame_imgPoint_XYZ: Optional[np.ndarray],
        totalBodyCOM_frame_XYZ: Optional[np.ndarray],
        rigid_bones_data: Optional[np.ndarray],
        processing_parameters: ProcessingParameterModel,
        queue: multiprocessing.Queue,
):
    if queue:
        handler = DirectQueueHandler(queue)
        handler.setFormatter(logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(handler)

    path_to_folder_where_we_will_save_this_data = processing_parameters.recording_info_model.output_data_folder_path

    logger.info("Saving post-processed data")
    save_numpy_array_to_disk(
        array_to_save=skel3d_frame_marker_xyz,
        file_name=f"{processing_parameters.tracking_model_info.name}_{DATA_3D_NPY_FILE_NAME}",
        save_directory=Path(path_to_folder_where_we_will_save_this_data),
    )

    if segment_COM_frame_imgPoint_XYZ is not None:
        logger.info("Saving segment center of mass data")
        save_numpy_array_to_disk(
            array_to_save=segment_COM_frame_imgPoint_XYZ,
            file_name=f"{processing_parameters.tracking_model_info.name}_{SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME}",
            save_directory=Path(path_to_folder_where_we_will_save_this_data) / CENTER_OF_MASS_FOLDER_NAME,
        )
    else:
        logger.debug("segment_COM_frame_imgPoint_XYZ is None, could not save")

    if totalBodyCOM_frame_XYZ is not None:
        logger.info("Saving total body center of mass data")
        save_numpy_array_to_disk(
            array_to_save=totalBodyCOM_frame_XYZ,
            file_name=f"{processing_parameters.tracking_model_info.name}_{TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME}",
            save_directory=Path(path_to_folder_where_we_will_save_this_data) / CENTER_OF_MASS_FOLDER_NAME,
        )
    else:
        logger.debug("totalBodyCOM_frame_XYZ is None, could not save")

    if rigid_bones_data is not None:
        logger.info("Saving rigid bones data")
        save_numpy_array_to_disk(
            array_to_save=rigid_bones_data,
            file_name=f"{processing_parameters.tracking_model_info.name}_{RIGID_BONES_NPY_FILE_NAME}",
            save_directory=path_to_folder_where_we_will_save_this_data,
        )
    else:
        logger.debug("rigid_bones_data is None, could not save")

    logger.info("Splitting and saving data...")
    split_and_save(
        skeleton_3d_data=skel3d_frame_marker_xyz,
        model_info=processing_parameters.tracking_model_info,
        output_data_folder_path=processing_parameters.recording_info_model.output_data_folder_path,
    )



from pathlib import Path
from typing import Optional, Union

from freemocap.system.paths_and_filenames.file_and_folder_names import (
    RAW_DATA_FOLDER_NAME,
    OUTPUT_DATA_FOLDER_NAME,
    CENTER_OF_MASS_FOLDER_NAME,
    RECORDING_SESSIONS_FOLDER_NAME,
)
from freemocap.system.paths_and_filenames.path_getters import get_freemocap_data_folder_path


def update_1_4_path_names(freemocap_data_folder_path: Optional[Union[str, Path]]) -> None:
    if freemocap_data_folder_path is None:
        freemocap_data_folder_path = get_freemocap_data_folder_path()

    print(f"Searching for recording folders in {freemocap_data_folder_path}")

    for session_folder in (Path(freemocap_data_folder_path) / RECORDING_SESSIONS_FOLDER_NAME).iterdir():
        if not session_folder.is_dir():
            continue

        for recording_folder in session_folder.iterdir():
            if not recording_folder.is_dir():
                continue

            update_recording_folder(recording_folder)


def update_recording_folder(recording_folder: Path) -> None:
    if (
            recording_folder.name == OUTPUT_DATA_FOLDER_NAME
    ):  # this covers cases where the recording folder isn't a subfolder of a session
        rename_skeleton_file_path(recording_folder)
        if (recording_folder / RAW_DATA_FOLDER_NAME).exists():
            rename_raw_data_paths(recording_folder / RAW_DATA_FOLDER_NAME)
        if (recording_folder / CENTER_OF_MASS_FOLDER_NAME).exists():
            rename_COM_paths(recording_folder / CENTER_OF_MASS_FOLDER_NAME)
    elif (recording_folder / OUTPUT_DATA_FOLDER_NAME).exists():
        rename_skeleton_file_path(recording_folder)
        if (recording_folder / OUTPUT_DATA_FOLDER_NAME / RAW_DATA_FOLDER_NAME).exists():
            rename_raw_data_paths(recording_folder / OUTPUT_DATA_FOLDER_NAME / RAW_DATA_FOLDER_NAME)
        if (recording_folder / OUTPUT_DATA_FOLDER_NAME / CENTER_OF_MASS_FOLDER_NAME).exists():
            rename_COM_paths(recording_folder / OUTPUT_DATA_FOLDER_NAME / CENTER_OF_MASS_FOLDER_NAME)


def rename_raw_data_paths(raw_data_folder_path: Path) -> None:
    print(f"renaming raw data files in {raw_data_folder_path}")
    for file in raw_data_folder_path.iterdir():
        if file.name == "mediapipe2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy":
            file.rename(file.parent / "mediapipe_2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy")
        elif file.name == "mediapipe3dData_numCams_numFrames_numTrackedPoints_reprojectionError.npy":
            file.rename(file.parent / "mediapipe_3dData_numCams_numFrames_numTrackedPoints_reprojectionError.npy")
        elif file.name == "mediapipe3dData_numFrames_numTrackedPoints_reprojectionError.npy":
            file.rename(file.parent / "mediapipe_3dData_numFrames_numTrackedPoints_reprojectionError.npy")
        elif file.name == "mediapipe3dData_numFrames_numTrackedPoints_spatialXYZ.npy":
            file.rename(file.parent / "mediapipe_3dData_numFrames_numTrackedPoints_spatialXYZ.npy")

    print(f"successfully renamed raw data files in {raw_data_folder_path}")


def rename_skeleton_file_path(output_data_folder_path: Path) -> None:
    print(f"renaming skeleton 3D files in {output_data_folder_path}")
    for file in output_data_folder_path.iterdir():
        if file.name == "mediaPipeSkel_3d_body_hands_face.npy":
            file.rename(file.parent / "mediapipe_skeleton_3d.npy")

    print(f"successfully renamed skeleton 3D files in {output_data_folder_path}")


def rename_COM_paths(COM_data_folder_path: Path) -> None:
    print(f"renaming center of mass files in {COM_data_folder_path}")

    for file in COM_data_folder_path.iterdir():
        if file.name == "segmentCOM_frame_joint_xyz.npy":
            file.rename(file.parent / "mediapipe_segmentCOM_frame_joint_xyz.npy")
        elif file.name == "total_body_center_of_mass_xyz.npy":
            file.rename(file.parent / "mediapipe_total_body_center_of_mass_xyz.npy")

    print(f"successfully renamed center of mass files in {COM_data_folder_path}")


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        update_1_4_path_names(Path(sys.argv[1]))
    else:
        update_1_4_path_names(None)



# __main__.py
import sys
from multiprocessing import freeze_support
from pathlib import Path

try:
    from freemocap.gui.qt.freemocap_main import qt_gui_main
except Exception:
    base_package_path = Path(__file__).parent.parent
    print(f"adding base_package_path: {base_package_path} : to sys.path")
    sys.path.insert(0, str(base_package_path))  # add parent directory to sys.path
    from freemocap.gui.qt.freemocap_main import qt_gui_main


def main():
    # set up so you can change the taskbar icon - https://stackoverflow.com/a/74531530/14662833
    import ctypes
    import freemocap

    if sys.platform == "win32":
        myappid = f"{freemocap.__package_name__}_{freemocap.__version__}"  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    qt_gui_main()


if __name__ == "__main__":
    freeze_support()
    print(f"Running `freemocap.__main__` from - {__file__}")

    main()



import logging
import multiprocessing
from pathlib import Path

import numpy as np
from skellytracker.process_folder_of_videos import process_folder_of_videos
from skellytracker.trackers.base_tracker.base_tracking_params import (
    BaseTrackingParams,
)
from skellytracker.trackers.base_tracker.model_info import ModelInfo

from freemocap.data_layer.recording_models.post_processing_parameter_models import (
    ProcessingParameterModel,
)
from freemocap.system.logging.configure_logging import log_view_logging_format_string
from freemocap.system.logging.queue_logger import DirectQueueHandler
from freemocap.system.paths_and_filenames.file_and_folder_names import (
    LOG_VIEW_PROGRESS_BAR_STRING,
    RAW_DATA_FOLDER_NAME,
)
from freemocap.tests.test_image_tracking_data_shape import test_image_tracking_data_shape

logger = logging.getLogger(__name__)


def run_image_tracking_pipeline(
    processing_parameters: ProcessingParameterModel,
    kill_event: multiprocessing.Event,
    queue: multiprocessing.Queue,
    use_tqdm: bool,
) -> np.ndarray:
    if queue:
        handler = DirectQueueHandler(queue)
        handler.setFormatter(logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(handler)

    if not processing_parameters.tracking_parameters_model.run_image_tracking:
        logger.info(
            f"Skipping 2d skeleton detection and loading data from: {processing_parameters.recording_info_model.data_2d_npy_file_path}"
        )
        try:
            image_data_numCams_numFrames_numTrackedPts_XYZ = np.load(
                processing_parameters.recording_info_model.data_2d_npy_file_path
            )
        except Exception as e:
            logger.error(e)
            raise RuntimeError("Failed to load 2D data, cannot continue processing") from e
    else:
        logger.info("Detecting 2d skeletons...")
        logger.info(LOG_VIEW_PROGRESS_BAR_STRING)
        # 2d skeleton detection
        image_data_numCams_numFrames_numTrackedPts_XYZ = run_image_tracking(
            model_info=processing_parameters.tracking_model_info,
            tracking_params=processing_parameters.tracking_parameters_model,
            synchronized_videos_folder_path=Path(
                processing_parameters.recording_info_model.synchronized_videos_folder_path
            ),
            output_data_folder_path=Path(processing_parameters.recording_info_model.output_data_folder_path)
            / RAW_DATA_FOLDER_NAME,
            kill_event=kill_event,
            use_tqdm=use_tqdm,
        )

    # TODO: check that skellytracker saved out data correctly
    try:
        test_image_tracking_data_shape(
            synchronized_video_folder_path=processing_parameters.recording_info_model.synchronized_videos_folder_path,
            image_tracking_data=image_data_numCams_numFrames_numTrackedPts_XYZ,
        )
    except (ValueError, AssertionError) as e:
        logger.error(e)
        raise ValueError("image tracking data is not valid, cannot continue processing")

    return image_data_numCams_numFrames_numTrackedPts_XYZ


def run_image_tracking(
    model_info: ModelInfo,
    tracking_params: BaseTrackingParams,
    synchronized_videos_folder_path: Path,
    output_data_folder_path: Path,
    kill_event: multiprocessing.Event = None,
    use_tqdm: bool = True,
):
    if hasattr(tracking_params, "use_yolo_crop_method") and tracking_params.use_yolo_crop_method:
        model_info.tracker_name = "YOLOMediapipeComboTracker"
    # TODO: this shouldn't be handled here

    image_data_numCams_numFrames_numTrackedPts_XYZ = process_folder_of_videos(
        model_info=model_info,
        tracking_params=tracking_params,
        synchronized_video_path=synchronized_videos_folder_path,
        output_folder_path=output_data_folder_path,
        annotated_video_path=None,
        num_processes=tracking_params.num_processes,
    )

    return image_data_numCams_numFrames_numTrackedPts_XYZ



import logging
import multiprocessing
from pathlib import Path
from typing import Union

import numpy as np
from skellyforge.freemocap_utils.config import default_settings
from skellyforge.freemocap_utils.constants import (
    TASK_FILTERING,
    PARAM_CUTOFF_FREQUENCY,
    PARAM_SAMPLING_RATE,
    PARAM_ORDER,
    PARAM_ROTATE_DATA,
    TASK_SKELETON_ROTATION,
    TASK_INTERPOLATION,
    TASK_FINDING_GOOD_FRAME,
)
from skellyforge.freemocap_utils.postprocessing_widgets.task_worker_thread import TaskWorkerThread
from skellytracker.trackers.base_tracker.model_info import ModelInfo

from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.system.logging.configure_logging import log_view_logging_format_string
from freemocap.system.logging.queue_logger import DirectQueueHandler
from freemocap.system.paths_and_filenames.file_and_folder_names import LOG_VIEW_PROGRESS_BAR_STRING

logger = logging.getLogger(__name__)


class PostProcessedDataHandler:
    def __init__(self):
        self.processed_skeleton = None

    def data_callback(self, processed_skeleton: np.ndarray):
        self.processed_skeleton = processed_skeleton


def save_numpy_array_to_disk(array_to_save: np.ndarray, file_name: str, save_directory: Union[str, Path]):
    if not file_name.endswith(".npy"):
        file_name += ".npy"
    Path(save_directory).mkdir(parents=True, exist_ok=True)
    np.save(
        str(Path(save_directory) / file_name),
        array_to_save,
    )


def get_settings_from_parameter_tree(recording_processing_parameter_model: ProcessingParameterModel):
    rec = recording_processing_parameter_model

    filter_sampling_rate = rec.post_processing_parameters_model.framerate
    filter_cutoff_frequency = rec.post_processing_parameters_model.butterworth_filter_parameters.cutoff_frequency
    filter_order = rec.post_processing_parameters_model.butterworth_filter_parameters.order
    run_butterworth_filter = rec.post_processing_parameters_model.run_butterworth_filter

    return filter_sampling_rate, filter_cutoff_frequency, filter_order, run_butterworth_filter


def adjust_default_settings(filter_sampling_rate, filter_cutoff_frequency, filter_order):
    adjusted_settings = default_settings.copy()

    adjusted_settings[TASK_FILTERING][PARAM_CUTOFF_FREQUENCY] = filter_cutoff_frequency
    adjusted_settings[TASK_FILTERING][PARAM_SAMPLING_RATE] = filter_sampling_rate
    adjusted_settings[TASK_FILTERING][PARAM_ORDER] = filter_order
    adjusted_settings[TASK_SKELETON_ROTATION][PARAM_ROTATE_DATA] = False

    return adjusted_settings


def run_post_processing_worker(
        raw_skel3d_frame_marker_xyz: np.ndarray,
        settings_dictionary: dict,
        landmark_names: list,
        run_butterworth_filter: bool,
) -> np.ndarray:
    def handle_thread_finished(results, post_processed_data_handler: PostProcessedDataHandler):
        # TODO: skellyforge should handle getting the final task results regardless of what was run
        if TASK_FILTERING in results and results[TASK_FILTERING] is not None:
            processed_skeleton = results[TASK_FILTERING]["result"]
        else:
            processed_skeleton = results[TASK_INTERPOLATION]["result"]
        post_processed_data_handler.data_callback(processed_skeleton=processed_skeleton)

    task_list = [TASK_INTERPOLATION, TASK_FILTERING, TASK_FINDING_GOOD_FRAME, TASK_SKELETON_ROTATION]

    if not run_butterworth_filter:
        task_list.remove(TASK_FILTERING)

    post_processed_data_handler = PostProcessedDataHandler()

    logger.info("Starting post-processing worker thread")
    logger.info(LOG_VIEW_PROGRESS_BAR_STRING)

    worker_thread = TaskWorkerThread(
        raw_skeleton_data=raw_skel3d_frame_marker_xyz,
        task_list=task_list,
        settings=settings_dictionary,
        all_tasks_finished_callback=lambda results: handle_thread_finished(results, post_processed_data_handler),
        landmark_names=landmark_names,
    )
    worker_thread.start()
    worker_thread.join()
    logger.info("Done with gap filling, filtering, and aligning")

    return post_processed_data_handler.processed_skeleton


def post_process_data(
        recording_processing_parameter_model: ProcessingParameterModel,
        raw_skel3d_frame_marker_xyz: np.ndarray,
        queue: multiprocessing.Queue,
) -> np.ndarray:
    if queue:
        handler = DirectQueueHandler(queue)
        handler.setFormatter(logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(handler)
    filter_sampling_rate, filter_cutoff_frequency, filter_order, run_butterworth_filter = (
        get_settings_from_parameter_tree(recording_processing_parameter_model)
    )
    adjusted_settings = adjust_default_settings(filter_sampling_rate, filter_cutoff_frequency, filter_order)
    processed_skeleton_array = run_post_processing_worker(
        raw_skel3d_frame_marker_xyz=raw_skel3d_frame_marker_xyz,
        settings_dictionary=adjusted_settings,
        landmark_names=get_landmark_names(model_info=recording_processing_parameter_model.tracking_model_info),
        run_butterworth_filter=run_butterworth_filter,
    )

    return processed_skeleton_array


def get_landmark_names(model_info: ModelInfo) -> list:
    if hasattr(model_info, "body_landmark_names"):
        return model_info.body_landmark_names
    elif hasattr(model_info, "landmark_names"):
        return model_info.landmark_names
    else:
        raise AttributeError("Model does not have landmark names")



import logging

from pydantic import BaseModel, ConfigDict
from skellytracker.trackers.base_tracker.base_tracking_params import BaseTrackingParams
from skellytracker.trackers.base_tracker.model_info import ModelInfo
from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import MediapipeTrackingParams, MediapipeModelInfo

from freemocap.data_layer.recording_models.recording_info_model import (
    RecordingInfoModel,
)

logger = logging.getLogger(__name__)


class AniposeTriangulate3DParametersModel(BaseModel):
    run_reprojection_error_filtering: bool = False
    reprojection_error_confidence_cutoff: float = 90
    minimum_cameras_to_reproject: int = 3
    use_triangulate_ransac_method: bool = False
    run_3d_triangulation: bool = True
    flatten_single_camera_data: bool = True


class ButterworthFilterParametersModel(BaseModel):
    sampling_rate: float = 30
    cutoff_frequency: float = 7
    order: int = 4


class PostProcessingParametersModel(BaseModel):
    framerate: float = 30.0
    butterworth_filter_parameters: ButterworthFilterParametersModel = ButterworthFilterParametersModel()
    max_gap_to_fill: int = 10
    run_butterworth_filter: bool = True


class ProcessingParameterModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    recording_info_model: RecordingInfoModel = None
    tracking_parameters_model: BaseTrackingParams = MediapipeTrackingParams()
    anipose_triangulate_3d_parameters_model: AniposeTriangulate3DParametersModel = AniposeTriangulate3DParametersModel()
    post_processing_parameters_model: PostProcessingParametersModel = PostProcessingParametersModel()
    tracking_model_info: ModelInfo = MediapipeModelInfo()



import json
import logging
from pathlib import Path
from typing import Union

from freemocap.utilities.download_sample_data import get_sample_data_path

logger = logging.getLogger(__name__)


def generate_jupyter_notebook(path_to_recording: Union[str, Path]):
    path_to_recording = Path(path_to_recording)
    recording_name = Path(path_to_recording).name
    path_to_template_notebook = Path(__file__).parent / "freemocap_template.ipynb"
    path_to_output_notebook = Path(path_to_recording) / f"{recording_name}.ipynb"

    with open(path_to_template_notebook, "r") as file:
        template_notebook = json.load(file)

    template_notebook["cells"][3]["source"] = [f'path_to_recording = "{str(path_to_recording)}"']

    with open(path_to_output_notebook, "w") as file:
        json.dump(template_notebook, file)


if __name__ == "__main__":
    path_to_recording = get_sample_data_path()
    # path_to_recording = r"PATH/TO/RECORDING/FOLDER" #specify path to a specific recording here (other wise it will just download sample data)

    generate_jupyter_notebook(path_to_recording)



import logging
import multiprocessing
from pathlib import Path
from typing import Optional

import numpy as np

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.get_anipose_calibration_object import (
    load_anipose_calibration_toml_from_path,
)
from freemocap.core_processes.capture_volume_calibration.by_camera_reprojection_filtering import (
    run_reprojection_error_filtering,
)
from freemocap.core_processes.capture_volume_calibration.save_3d_data_to_npy import (
    save_3d_data_to_npy,
)
from freemocap.core_processes.capture_volume_calibration.triangulate_3d_data import triangulate_3d_data
from freemocap.core_processes.post_process_skeleton_data.process_single_camera_skeleton_data import (
    process_single_camera_skeleton_data,
)
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.system.logging.configure_logging import log_view_logging_format_string
from freemocap.system.logging.queue_logger import DirectQueueHandler
from freemocap.system.paths_and_filenames.file_and_folder_names import LOG_VIEW_PROGRESS_BAR_STRING

logger = logging.getLogger(__name__)


def get_triangulated_data(
        image_data_numCams_numFrames_numTrackedPts_XYZ: np.ndarray,
        processing_parameters: ProcessingParameterModel,
        kill_event: Optional[multiprocessing.Event] = None,
        queue: Optional[multiprocessing.Queue] = None,
) -> np.ndarray:
    if queue:
        handler = DirectQueueHandler(queue)
        handler.setFormatter(logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(handler)

    if image_data_numCams_numFrames_numTrackedPts_XYZ.shape[0] == 1:
        logger.info("Skipping 3d triangulation for single camera data")
        logger.info(LOG_VIEW_PROGRESS_BAR_STRING)

        (skel3d_frame_marker_xyz, skeleton_reprojection_error_fr_mar) = process_single_camera_skeleton_data(
            input_image_data_frame_marker_xyz=image_data_numCams_numFrames_numTrackedPts_XYZ[0],
            raw_data_folder_path=Path(processing_parameters.recording_info_model.raw_data_folder_path),
            file_prefix=processing_parameters.tracking_model_info.name,
            project_to_z_plane=processing_parameters.anipose_triangulate_3d_parameters_model.flatten_single_camera_data,
        )
    elif not processing_parameters.anipose_triangulate_3d_parameters_model.run_3d_triangulation:
        logger.info(
            f"Skipping 3d triangulation and loading data from: {processing_parameters.recording_info_model.raw_data_3d_npy_file_path}"
        )
        logger.info(LOG_VIEW_PROGRESS_BAR_STRING)
        try:
            skel3d_frame_marker_xyz = np.load(
                processing_parameters.recording_info_model.raw_data_3d_npy_file_path
            )
            skeleton_reprojection_error_fr_mar = np.load(
                processing_parameters.recording_info_model.reprojection_error_data_npy_file_path
            )
        except Exception as e:
            logger.error(e)
            raise RuntimeError("Failed to load 3D data, cannot continue processing") from e
    else:
        logger.info("Triangulating 3d skeletons...")
        logger.info(LOG_VIEW_PROGRESS_BAR_STRING)

        if not processing_parameters.recording_info_model.calibration_toml_check:
            raise FileNotFoundError(
                f"No calibration file found at: {processing_parameters.recording_info_model.calibration_toml_path}"
            )

        anipose_calibration_object = load_anipose_calibration_toml_from_path(
            camera_calibration_data_toml_path=processing_parameters.recording_info_model.calibration_toml_path,
            save_copy_of_calibration_to_this_path=processing_parameters.recording_info_model.path,
        )
        (
            skel3d_frame_marker_xyz,
            skeleton_reprojection_error_fr_mar,
            skeleton_reprojection_error_cam_fr_mar,
        ) = triangulate_3d_data(
            anipose_calibration_object=anipose_calibration_object,
            image_2d_data=image_data_numCams_numFrames_numTrackedPts_XYZ[:, :, :, :2],
            use_triangulate_ransac=processing_parameters.anipose_triangulate_3d_parameters_model.use_triangulate_ransac_method,
            kill_event=kill_event,
        )
        save_3d_data_to_npy(
            data3d_numFrames_numTrackedPoints_XYZ=skel3d_frame_marker_xyz,
            data3d_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_fr_mar,
            data3d_numCams_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_cam_fr_mar,
            path_to_folder_where_data_will_be_saved=processing_parameters.recording_info_model.raw_data_folder_path,
            processing_level="raw",
            file_prefix=processing_parameters.tracking_model_info.name,
        )

        if processing_parameters.anipose_triangulate_3d_parameters_model.run_reprojection_error_filtering:
            logger.info("Filtering 3d triangulation...")
            skel3d_frame_marker_xyz = run_reprojection_error_filtering(
                image_data_numCams_numFrames_numTrackedPts_XYZ=image_data_numCams_numFrames_numTrackedPts_XYZ,
                raw_skel3d_frame_marker_xyz=skel3d_frame_marker_xyz,
                skeleton_reprojection_error_fr_mar=skeleton_reprojection_error_fr_mar,
                skeleton_reprojection_error_cam_fr_mar=skeleton_reprojection_error_cam_fr_mar,
                anipose_calibration_object=anipose_calibration_object,
                processing_parameters=processing_parameters,
            )

    return skel3d_frame_marker_xyz



from pathlib import Path

import numpy as np
from pydantic import BaseModel
from skellytracker.process_folder_of_videos import process_folder_of_videos
from skellytracker.trackers.charuco_tracker.charuco_model_info import (
    CharucoModelInfo, CharucoTrackingParams)


class CharucoNeighborStats(BaseModel):
    mean_distance: float
    median_distance: float
    std_distance: float
    mean_error: float


def get_charuco_2d_data(calibration_videos_folder_path: str | Path, num_processes: int = 1):
    calibration_videos_folder_path = Path(calibration_videos_folder_path)
    return process_folder_of_videos(
        model_info=CharucoModelInfo(),
        tracking_params=CharucoTrackingParams(),
        synchronized_video_path=Path(calibration_videos_folder_path),
        num_processes=num_processes,
    )


def get_charuco_neighbor_pairs(rows: int, cols: int):
    neighbors = []
    for row in range(rows):
        for col in range(cols):
            idx = row * cols + col
            if col < cols - 1:
                neighbors.append((idx, idx + 1))  # right neighbor
            if row < rows - 1:
                neighbors.append((idx, idx + cols))  # bottom neighbor
    return neighbors


def get_neighbor_distances(
        number_of_squares_width: int, number_of_squares_height: int, charuco_3d_data: np.ndarray
) -> list:
    distances_per_frame = []
    num_cols = number_of_squares_width - 1
    num_rows = number_of_squares_height - 1

    neighbor_pairs = get_charuco_neighbor_pairs(num_rows, num_cols)

    for frame_num in range(charuco_3d_data.shape[0]):
        points_3d = charuco_3d_data[frame_num]

        # skip frames with missing points
        if np.isnan(points_3d).any():
            continue

        marker_distances = []
        for i, j in neighbor_pairs:
            pt1, pt2 = points_3d[i], points_3d[j]
            if not np.isnan(pt1).any() and not np.isnan(pt2).any():
                distance = np.linalg.norm(pt2 - pt1)
                marker_distances.append(distance)

        if marker_distances:
            distances_per_frame.append(marker_distances)
    return np.array(distances_per_frame, dtype=np.float64)


def get_neighbor_stats(distances: np.ndarray, charuco_square_size_mm: float) -> dict:
    """Computes statistics for the distances between neighboring points"""
    mean_distance = np.nanmean(distances)
    median_distance = np.nanmedian(distances)
    std_distance = np.nanstd(distances)
    mean_error = mean_distance - charuco_square_size_mm

    return CharucoNeighborStats(
        mean_distance=mean_distance, median_distance=median_distance, std_distance=std_distance, mean_error=mean_error
    )


def calculate_calibration_diagnostics(
        charuco_3d_data: np.ndarray,
        charuco_square_size_mm: float,
        number_of_squares_width: int,
        number_of_squares_height: int,
) -> CharucoNeighborStats:
    distances = get_neighbor_distances(
        number_of_squares_width=number_of_squares_width,
        number_of_squares_height=number_of_squares_height,
        charuco_3d_data=charuco_3d_data,
    )

    stats = get_neighbor_stats(distances=distances, charuco_square_size_mm=charuco_square_size_mm)
    return stats



from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QLabel, QPushButton, QVBoxLayout, QSizePolicy


class GroundPlaneCalibrationFailedDialog(QDialog):
    def __init__(self, message: str):
        super().__init__()
        self.setWindowTitle("Ground Plane Calibration Failed")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(500, 200)
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.WindowStaysOnTopHint)

        # Entire dialog styling (includes title bar area)
        self.setStyleSheet(
            """
            QDialog {
                background-color: #ffe0dc;
            }
            QLabel#titleLabel {
                font-weight: bold;
                font-size: 12pt;
                color: black;
            }
            QLabel#messageLabel {
                font-size: 10.5pt;
                color: black;
            }
        """
        )

        layout = QVBoxLayout()
        layout.setContentsMargins(20, 20, 20, 20)
        layout.setSpacing(15)

        # Title
        title = QLabel("Ground Plane Calibration Failed")
        title.setObjectName("titleLabel")

        # Message
        message_label = QLabel(message)
        message_label.setObjectName("messageLabel")
        message_label.setWordWrap(True)
        message_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        # OK Button
        ok_button = QPushButton("OK")
        ok_button.clicked.connect(self.accept)

        layout.addWidget(title)
        layout.addWidget(message_label)
        layout.addStretch(1)
        layout.addWidget(ok_button, alignment=Qt.AlignmentFlag.AlignRight)

        self.setLayout(layout)



import threading
from pathlib import Path

from PySide6.QtWidgets import QVBoxLayout, QDialog, QPushButton, QLabel, QHBoxLayout, QFileDialog

from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState, save_gui_state
from freemocap.system.paths_and_filenames.path_getters import get_gui_state_json_path


class SetDataFolderDialog(QDialog):
    def __init__(self, gui_state: GuiState, kill_thread_event: threading.Event, parent=None) -> None:
        super().__init__(parent=parent)
        self.gui_state = gui_state
        self.kill_thread_event = kill_thread_event

        self._initUI()

    def _initUI(self) -> None:
        self.setMinimumSize(500, 200)

        self.setWindowTitle("Set Freemocap Data Folder")

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._title_message_label = QLabel(f"Freemocap Data Folder: {self.gui_state.freemocap_data_folder_path}")
        self._title_message_label.setWordWrap(True)
        self._layout.addWidget(self._title_message_label)

        self._change_data_folder_button = QPushButton("Change Folder Location")
        self._change_data_folder_button.clicked.connect(self._change_data_folder)
        self._layout.addWidget(self._change_data_folder_button)

        self._reboot_warning_label = QLabel("Changing the freemocap data folder will reboot the GUI.")
        self._layout.addWidget(self._reboot_warning_label)

        button_box = QHBoxLayout()

        _cancel_button = QPushButton("Cancel")
        _cancel_button.clicked.connect(self.reject)
        button_box.addWidget(_cancel_button)

        self._done_button = QPushButton("Save and Reboot")
        self._done_button.clicked.connect(self._save_and_accept)
        button_box.addWidget(self._done_button)

        self._layout.addLayout(button_box)

    def _change_data_folder(self) -> None:
        self.new_folder_path = QFileDialog.getExistingDirectory(
            self,
            "Select New Data Folder",
            str(self.gui_state.freemocap_data_folder_path),
            QFileDialog.ShowDirsOnly,
        )
        self._title_message_label.setText(f"Freemocap Data Folder: {self.new_folder_path}")

        if self.new_folder_path == "":
            self._done_button.setEnabled(False)
        else:
            self._done_button.setEnabled(True)

    def _save_and_accept(self) -> None:
        if self.new_folder_path is not None:
            freemocap_data_folder_path = Path(self.new_folder_path)
            self.gui_state.freemocap_data_folder_path = str(freemocap_data_folder_path)
            save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())

            self.accept()



import logging
import multiprocessing
from pathlib import Path

import numpy as np

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.freemocap_anipose import CameraGroup
from freemocap.core_processes.capture_volume_calibration.save_3d_data_to_npy import (
    save_3d_data_to_npy,
)

logger = logging.getLogger(__name__)


def triangulate_3d_data(
        anipose_calibration_object: CameraGroup,
        image_2d_data: np.ndarray,
        use_triangulate_ransac: bool = False,
        kill_event: multiprocessing.Event = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    number_of_cameras = image_2d_data.shape[0]
    number_of_frames = image_2d_data.shape[1]
    number_of_tracked_points = image_2d_data.shape[2]
    number_of_spatial_dimensions = image_2d_data.shape[3]

    if not number_of_spatial_dimensions == 2:
        logger.error(
            f"This is supposed to be 2D data but, number_of_spatial_dimensions: {number_of_spatial_dimensions}"
        )
        raise ValueError

    # reshape data to collapse across 'frames' so it becomes [number_of_cameras,
    # number_of_2d_points(numFrames*numPoints), XY]
    data2d_flat = image_2d_data.reshape(number_of_cameras, -1, 2)

    logger.info(
        f"Reconstructing 3d points from 2d points with shape: \n"
        f"number_of_cameras: {number_of_cameras},\n"
        f"number_of_frames: {number_of_frames}, \n"
        f"number_of_tracked_points: {number_of_tracked_points},\n"
        f"number_of_spatial_dimensions: {number_of_spatial_dimensions}"
    )

    if use_triangulate_ransac:
        logger.info("Using `triangulate_ransac` method")
        data3d_flat = anipose_calibration_object.triangulate_ransac(data2d_flat, progress=True, kill_event=kill_event)
    else:
        logger.info("Using simple `triangulate` method ")
        data3d_flat = anipose_calibration_object.triangulate(data2d_flat, progress=True, kill_event=kill_event)

    spatial_data3d_numFrames_numTrackedPoints_XYZ = data3d_flat.reshape(number_of_frames, number_of_tracked_points, 3)

    data3d_reprojectionError_flat = anipose_calibration_object.reprojection_error(data3d_flat, data2d_flat, mean=True)
    data3d_reprojectionError_full = anipose_calibration_object.reprojection_error(data3d_flat, data2d_flat, mean=False)
    reprojectionError_cam_frame_marker = np.linalg.norm(data3d_reprojectionError_full, axis=2).reshape(
        number_of_cameras, number_of_frames, number_of_tracked_points
    )

    reprojection_error_data3d_numFrames_numTrackedPoints = data3d_reprojectionError_flat.reshape(
        number_of_frames, number_of_tracked_points
    )

    return (
        spatial_data3d_numFrames_numTrackedPoints_XYZ,
        reprojection_error_data3d_numFrames_numTrackedPoints,
        reprojectionError_cam_frame_marker,
    )


if __name__ == "__main__":
    from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.get_anipose_calibration_object import (
        load_anipose_calibration_toml_from_path,
    )
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_2d_path",
        type=str,
        help="path to 2d data",
        required=True,
    )
    parser.add_argument(
        "--output_data_folder_path",
        type=str,
        help="path to output folder",
        required=False,
    )
    parser.add_argument(
        "--calibration_file_path",
        type=str,
        help="path to calibration file",
        required=False,
    )
    parser.add_argument(
        "--save_data_as_csv",
        type=bool,
        help="save data as csv",
        required=False,
    )
    parser.add_argument(
        "--use_triangulate_ransac",
        type=bool,
        help="use triangulate ransac anipose method ",
        required=False,
    )
    parser.add_argument(
        "--file_prefix",
        type=str,
        help="file prefix",
        required=False,
    )
    args = parser.parse_args()

    data_2d = np.load(args.data_2d_path)

    if args.output_data_folder_path is None:
        args.output_data_folder_path = Path(args.data_2d_path).parent

    if args.save_data_as_csv is None:
        args.save_data_as_csv = True  # default

    if args.use_triangulate_ransac is None:
        args.use_triangulate_ransac = True

    if args.file_prefix is None:
        args.file_prefix = ""

    if args.calibration_file_path:
        anipose_calibration_object = load_anipose_calibration_toml_from_path(args.calibration_file_path)
    else:
        anipose_calibration_object = load_anipose_calibration_toml_from_path(
            Path(args.data_2d_path).parent.parent / "camera_calibration_data.toml"
        )

    (
        skel3d_frame_marker_xyz,
        skeleton_reprojection_error_fr_mar,
        skeleton_reprojection_error_cam_fr_mar,
    ) = triangulate_3d_data(
        anipose_calibration_object=anipose_calibration_object,
        image_2d_data=data_2d,
        use_triangulate_ransac=args.use_triangulate_ransac,
    )
    save_3d_data_to_npy(
        data3d_numFrames_numTrackedPoints_XYZ=skel3d_frame_marker_xyz,
        data3d_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_fr_mar,
        data3d_numCams_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_cam_fr_mar,
        path_to_folder_where_data_will_be_saved=args.output_data_folder_path,
        processing_level="raw",
        file_prefix=args.file_prefix,
    )



import logging
import uuid

import requests

logger = logging.getLogger(__name__)


class PipedreamPings:
    def __init__(
            self,
    ):
        self._pipedream_url = "https://eowipinr6rcpbo0.m.pipedream.net"
        self._pings_dict = {}
        self.update_pings_dict(key="session_uuid", value=str(uuid.uuid4()))

    def update_pings_dict(self, key, value):
        self._pings_dict[key] = value

    def send_pipedream_ping(self):
        logger.info(
            f"Sending anonymous ping to pipedream to let the devs know that someone is using this, which will allow "
            f"us to secure more funding to support this project. This is the dictionary of info we're sending: \n "
            f"{self._pings_dict} "
        )

        try:
            requests.post(self._pipedream_url, json=self._pings_dict, timeout=(5, 60))
        except requests.RequestException:
            logger.error("Failed to send anonymous ping to pipedream")



import logging
import multiprocessing
import time
from pathlib import Path

from PySide6.QtCore import Signal, QThread

from freemocap.core_processes.process_motion_capture_videos.process_recording_folder import process_recording_folder
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.system.paths_and_filenames.file_and_folder_names import RECORDING_PARAMETERS_JSON_FILE_NAME
from freemocap.utilities.save_dictionary_to_json import save_dictionary_to_json

logger = logging.getLogger(__name__)


class ProcessMotionCaptureDataThreadWorker(QThread):
    finished = Signal(bool)
    in_progress = Signal(object)

    def __init__(
            self, post_processing_parameters: ProcessingParameterModel, kill_event: multiprocessing.Event, parent=None
    ):
        super().__init__()
        self._processing_parameters = post_processing_parameters
        self._kill_event = kill_event

        self._queue = multiprocessing.Queue()

        self._process = multiprocessing.Process(
            target=process_recording_folder, args=(self._processing_parameters, self._kill_event, self._queue)
        )
        self._success = None

    def run(self):
        logger.info(
            f"Beginning processing of motion capture data with parameters: {self._processing_parameters.model_dump(exclude={'tracking_model_info'})}"
        )
        self._kill_event.clear()

        recording_info_dict = self._processing_parameters.model_dump(exclude={"recording_info_model"})
        Path(self._processing_parameters.recording_info_model.output_data_folder_path).mkdir(
            parents=True, exist_ok=True
        )

        save_dictionary_to_json(
            save_path=self._processing_parameters.recording_info_model.output_data_folder_path,
            file_name=RECORDING_PARAMETERS_JSON_FILE_NAME,
            dictionary=recording_info_dict,
        )

        try:
            self._process.start()
            while self._process.is_alive():
                time.sleep(0.01)
                if self._queue.empty():
                    continue
                else:
                    record = self._queue.get()
                    print(f"message: {record.msg}")
                    self.in_progress.emit(record)

            while not self._queue.empty():
                record = self._queue.get()
                print(f"message: {record.msg}")
                self.in_progress.emit(record)

            self._success = True
            logger.info("Finished processing session folder!")

        except Exception as e:  # noqa
            record = self._queue.get()
            logger.error(f"Error processing session folder: {str(record)}")
            self._success = False

        self.finished.emit(self._success)



from pathlib import Path
from typing import Union

from PySide6.QtWidgets import QWidget


def apply_css_style_sheet(qt_widget: QWidget, path_to_css_file: Union[str, Path]):
    with open(path_to_css_file, "r") as css_file:
        css_string = css_file.read()
        qt_widget.setStyleSheet(css_string)
